from pydantic import BaseModel
from helix.data.read_type import ReadType


class AlignmentStats(BaseModel):
    skipped_samples: int = None
    samples_count: int = None
    sequencer: str = None
    duplicate: int = None
    count_length: int = None
    average_length: float = None
    standard_dev_length: float = None
    count_insert_size: int = None
    average_insert_size: float = None
    standard_dev_insert_size: float = None
    count_quality: int = None
    average_quality: float = None
    standard_dev_quality: float = None
    read_type: ReadType = None



import enum

from PySide6.QtWidgets import QGroupBox, QRadioButton, QVBoxLayout, QWidget

from helix.data.file_type import FileType


class ExtractTargetFormat(enum.Enum):
    Microarray = enum.auto()
    SAM = enum.auto()
    BAM = enum.auto()
    CRAM = enum.auto()
    FASTA = enum.auto()
    FASTQ = enum.auto()
    HTML = enum.auto()
    VCF = enum.auto()
    Unknown = enum.auto()


class FormatSelection(QWidget):
    def __init__(self, parent, file_type: FileType):
        super().__init__(parent)
        self.file_type = ExtractTargetFormat[file_type.name]
        self.format = QGroupBox("Select the destination format", self)
        self.inner_layout = QVBoxLayout(self.format)

        self.setObjectName(FormatSelection.__name__)
        self.inner_layout.setObjectName("groupBoxLayout")

        self._format_options = [
            QRadioButton(label, self.format)
            for label in [
                "Microarray, textual format containing genotyping info",
                "SAM, textual format for aligned file",
                "BAM, binary equivalent of SAM",
                "CRAM, highly compressed equivalent of BAM",
                "FASTA, textual format obtained with a consensus algorithm",
                "FASTQ, textual format for unaligned files",
                "HTML, report containing an overview of the file",
                "VCF, textual format for variants",
            ]
        ]

        self._format_options[0].setObjectName(ExtractTargetFormat.Microarray.name)
        self._format_options[1].setObjectName(ExtractTargetFormat.SAM.name)
        self._format_options[2].setObjectName(ExtractTargetFormat.BAM.name)
        self._format_options[3].setObjectName(ExtractTargetFormat.CRAM.name)
        self._format_options[4].setObjectName(ExtractTargetFormat.FASTA.name)
        self._format_options[5].setObjectName(ExtractTargetFormat.FASTQ.name)
        self._format_options[6].setObjectName(ExtractTargetFormat.HTML.name)
        self._format_options[7].setObjectName(ExtractTargetFormat.VCF.name)

        self.options_by_type = {
            ExtractTargetFormat[x.objectName()]: x for x in self._format_options
        }
        self.options_by_type[self.file_type].hide()

        for item in self._format_options:
            self.inner_layout.addWidget(item)

        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.format)
        self.setLayout(self.main_layout)
        self.hide()



import logging
import typing
from collections import OrderedDict
from pathlib import Path

from helix.data.alignment_map.alignment_map_header_metadata import (
    AlignmentMapHeaderMetadata,
)
from helix.data.alignment_map.alignment_map_header_program import (
    AlignmentMapHeaderProgram,
)
from helix.data.alignment_map.alignment_map_header_read_group import (
    AlignmentMapHeaderReadGroup,
)
from helix.data.alignment_map.alignment_map_header_sequence import (
    AlignmentMapHeaderSequence,
)
from helix.data.chromosome_name_type import ChromosomeNameType
from helix.data.mitochondrial_name_type import MitochondrialNameType
from helix.data.sorting import Sorting


class AlignmentMapHeader:
    """Parse the output from `samtools view -H`"""

    def __init__(self, lines) -> None:
        self.metadata = None
        self.sequences: OrderedDict[str, AlignmentMapHeaderSequence] = OrderedDict()
        self.programs: list[AlignmentMapHeaderProgram] = []
        self.read_groups: list[AlignmentMapHeaderReadGroup] = []
        self.comments: list[str] = []

        self._handlers = {
            "@HD": self._metadata_process,
            "@SQ": self._sequence_process,
            "@PG": self._program_process,
            "@CO": self._comment_process,
            "@RG": self._read_group_process,
        }
        self._load(lines)

    def _program_process(self, parts: list[str]):
        program = AlignmentMapHeaderProgram()
        for part in parts:
            if part.startswith("ID"):
                program.id = part.split(":", 1)[1]
            elif part.startswith("PN"):
                program.id = part.split(":", 1)[1]
            elif part.startswith("CL"):
                program.id = part.split(":", 1)[1]
            elif part.startswith("PP"):
                program.previous = part.split(":", 1)[1]
                if program.previous in [x.id for x in self.programs]:
                    prev = [x for x in self.programs if x.id == program.previous]
                    program.previous = prev[0]
            elif part.startswith("DS"):
                program.description = part.split(":", 1)[1]
            elif part.startswith("VN"):
                program.program_version = part.split(":", 1)[1]
        self.programs.append(program)

    def _read_group_process(self, parts: list[str]):
        read_group = AlignmentMapHeaderReadGroup()
        for part in parts:
            if part.startswith("ID"):
                read_group.id = part.split(":", 1)[1]
            elif part.startswith("BC"):
                read_group.barcode = part.split(":", 1)[1]
            elif part.startswith("CN"):
                read_group.sequencing_center = part.split(":", 1)[1]
            elif part.startswith("DS"):
                read_group.description = part.split(":", 1)[1]
            elif part.startswith("DT"):
                read_group.date = part.split(":", 1)[1]
            elif part.startswith("FO"):
                read_group.flow_order = part.split(":", 1)[1]
            elif part.startswith("KS"):
                read_group.key_sequence = part.split(":", 1)[1]
            elif part.startswith("LB"):
                read_group.library = part.split(":", 1)[1]
            elif part.startswith("PG"):
                read_group.programs = part.split(":", 1)[1]
            elif part.startswith("PI"):
                read_group.predicted_median_insert_size = part.split(":", 1)[1]
            elif part.startswith("PL"):
                read_group.platform = part.split(":", 1)[1]
            elif part.startswith("PM"):
                read_group.platform_model = part.split(":", 1)[1]
            elif part.startswith("PU"):
                read_group.platform_unit = part.split(":", 1)[1]
            elif part.startswith("SM"):
                read_group.sample = part.split(":", 1)[1]
        self.read_groups.append(read_group)

    def _comment_process(self, parts: list[str]):
        self.comments.append(" ".join(parts))
        return

    def _metadata_process(self, parts: list[str]):
        metadata = AlignmentMapHeaderMetadata()
        for part in parts:
            if part.startswith("VN"):
                metadata.version = part.split(":", 1)[1]
            elif part.startswith("SO:"):
                metadata.sorted = self._sorted(part)
            elif part.startswith("GO:"):
                metadata.grouping = part.split(":", 1)[1]
            elif part.startswith("SS:"):
                metadata.subsorting = part.split(":", 1)[1]
        self.metadata = metadata

    def _sequence_process(self, parts: typing.List[str]):
        entry = AlignmentMapHeaderSequence()

        for part in parts:
            if part.startswith("SN"):
                entry.name = part.split(":", 1)[1]
            elif part.startswith("LN"):
                entry.length = int(part.split(":", 1)[1])
            elif part.startswith("M5"):
                entry.md5 = part.split(":", 1)[1]
            elif part.startswith("UR"):
                entry.uri = part.split(":", 1)[1]
            elif part.startswith("AH"):
                entry.alternate_locus = part.split(":", 1)[1]
            elif part.startswith("AN"):
                entry.alternative_names = part.split(":", 1)[1].split(",")
            elif part.startswith("AS"):
                entry.genome_assembly_identifier = part.split(":", 1)[1]
            elif part.startswith("DS"):
                entry.description = part.split(":", 1)[1]
            elif part.startswith("SP"):
                entry.species = part.split(":", 1)[1]
            elif part.startswith("TP"):
                entry.molecule_topology = part.split(":", 1)[1]
        if entry.name is None:
            raise RuntimeError("Unable to find the name of the sequence.")
        if entry.length is None:
            raise RuntimeError("Unable to find the length of the sequence.")

        self.sequences[entry.name] = entry

    def _load(self, lines):
        for index, line in enumerate(lines):
            line = line.strip()
            if line == "":
                continue
            parts = line.split("\t")
            line_type = parts[0].strip()
            if line_type in self._handlers:
                self._handlers[line_type](parts)
            else:
                logging.debug(
                    f"Skipping line {index} in dictionary because unrecognized type {line_type}."
                )

    def load_from_file(path: Path) -> "AlignmentMapHeader":
        if path is None:
            raise ValueError("path cannot be None.")
        if not path.exists():
            raise FileNotFoundError(f"Unable to find file: {str(path)}")

        with path.open("rt") as f:
            return AlignmentMapHeader(f)

    def _sorted(self, value: str) -> Sorting:
        value = value.split(":")[1].strip().lower()
        if "coordinate" in value:
            return Sorting.Coordinate
        elif "unsorted" in value:
            return Sorting.Unsorted
        elif "unknown" in value:
            return Sorting.Unknown
        elif "queryname" in value:
            return Sorting.QueryName
        logging.debug(f"Unable to recognize the sorting header {value}")
        return Sorting.Unknown

    def chromosome_name_type(self):
        patterns_type = {
            ChromosomeNameType.GenBank: ["CM0", "CP", "J0", "NC_"],
            ChromosomeNameType.Chr: ["chr"],
            ChromosomeNameType.Number: ["1"],
        }

        for type, patterns in patterns_type.items():
            for sequence in self.sequences.keys():
                if any([x in sequence for x in patterns]):
                    return type
        return ChromosomeNameType.Unknown

    def mtdna_name_type(self):
        patterns_type = {
            MitochondrialNameType.chrMT: ["chrMT"],
            MitochondrialNameType.MT: ["MT"],
            MitochondrialNameType.chrM: ["chrM"],
            MitochondrialNameType.M: ["M"],
        }

        for type, patterns in patterns_type.items():
            if any([x in self.sequences.keys() for x in patterns]):
                return type

        if self.chromosome_name_type() == ChromosomeNameType.GenBank:
            return MitochondrialNameType.Accession
        return MitochondrialNameType.Unknown

    def sequence_count(self):
        return len(self.sequences)



import enum
import logging
from collections import OrderedDict

from helix.data.genome import Genome
from helix.data.sequence import Sequence


class ReferenceStatus(enum.Enum):
    Available = enum.auto()
    Downloadable = enum.auto()
    Buildable = enum.auto()
    Ambiguous = enum.auto()
    Unknown = enum.auto()


class Reference:
    """This class represent a reference genome."""

    def __init__(
        self,
        reference_map: OrderedDict[Sequence, list[Sequence]],
        logger=logging.getLogger(__name__),
    ):
        self._logger = logger
        self.reference_map = reference_map
        self._genome_map = self._index_by_genome()
        self.matching: list[Genome] = self._get_matching_genomes()
        sorted = list([x for x in self._genome_map.keys() if x is not None])
        sorted.sort(
            key=lambda x: len(
                [y for y in self._genome_map[x].values() if y is not None]
            )
        )
        self.build = []
        if len(self.matching) > 0:
            self.build = set(x.build for x in self.matching)
        elif len(sorted) > 0:
            self.build = [f"Likely {sorted[-1].build}"]

        if len(self.build) > 1:
            self._logger.warn(
                "Found more than one valid builds for the reference of the file. "
                + "This is likely an issue with the reference genome metadata. "
                + "Please open a bug report."
            )
        self.build = " ".join(self.build)
        self.ready_reference = None
        for match in self.matching:
            if match.fasta.exists():
                self.ready_reference = match
                break
        if self.ready_reference is not None:
            self.status = ReferenceStatus.Available
        elif len(self.matching) > 0:
            self.status = ReferenceStatus.Downloadable
        # There's a possibility to build it only if every sequence has a
        # match in the metadata.
        elif all([True if len(x) > 0 else False for x in self.reference_map.values()]):
            self.status = ReferenceStatus.Buildable
        else:
            self.status = ReferenceStatus.Unknown

    def _md5_available(self):
        """Return True if MD5 is available for every input sequence.

        Returns:
            bool: True if MD5 available.
        """
        return all([x.md5 is not None for x in self.reference_map.keys()])

    def _index_by_genome(self) -> dict[Genome, dict[Sequence, Sequence]]:
        """For every genome that has at least one match in the input sequence,
        return a map between its sequences and input sequences.

        Returns:
            dict[Genome, dict[Sequence, Sequence]]: Map Genome <-> Sequences.
        """
        parents: set[Genome] = set(
            x.parent for y in self.reference_map.values() for x in y
        )
        genome_map = {x: {y: None for y in x.sequences} for x in parents}
        for parent in parents:
            genome_map[parent][None] = []
            for sequence in parent.sequences:
                for query_sequence, match_sequences in self.reference_map.items():
                    if sequence in match_sequences:
                        genome_map[parent][sequence] = query_sequence
                        break
            for query_sequence in self.reference_map.keys():
                if query_sequence not in genome_map[parent].values():
                    genome_map[parent][None].append(query_sequence)
        return genome_map

    def _get_matching_genomes(self):
        """Return a list of genomes that are guaranteed to match the input sequence
        by name, length or MD5 (if available in the input sequence).

        Returns:
            list[Genome]: list of matching genomes.
        """
        match_list = []
        # None is a special key for all the sequences
        # that have 0 matches among all the known references.
        # If None is in _genome_map means we didn't find a
        # reference.
        for genome, matches in self._genome_map.items():
            if len(matches[None]) > 0:
                continue
            matching = True
            for genome_sequence, query_sequence in matches.items():
                if genome_sequence is None:
                    continue
                if query_sequence is None:
                    matching = False
                    break
                if genome_sequence.name != query_sequence.name:
                    matching = False
                    break
                if query_sequence.md5 is not None:
                    if genome_sequence.md5 != query_sequence.md5:
                        matching = False
                        break
                if genome_sequence.length != query_sequence.length:
                    matching = False
                    break
            if matching:
                self._logger.info(f"{genome!s} is a perfect match.")
                match_list.append(genome)
        return match_list

    def __str__(self) -> str:
        return f"{self.build} ({self.status.name})"

    def __repr__(self) -> str:
        return self.__str__()



import logging
import shutil
import time
from collections import OrderedDict
from pathlib import Path

from helix.configuration import MANAGER_CFG
from helix.data.microarray_converter import MicroarrayConverterTarget
from helix.microarray.microarray_line_formatter import TARGET_FORMATTER_MAP
from helix.microarray.raw_file import RawEntry, RawFile
from helix.naming.converter import Converter
from helix.naming.orderer import SequenceOrderer


class MicroarrayConverter:
    def __init__(self, config=MANAGER_CFG.REPOSITORY) -> None:
        self._config = config
        self._template_folder = self._config.metadata.joinpath("microarray_templates")
        self._head_folder = self._template_folder.joinpath("head")
        self._body_folder = self._template_folder.joinpath("body")

        if not self._body_folder.exists():
            raise FileNotFoundError(
                f"Unable to find microarray template body folder at {self._body_folder!s}"
            )
        if not self._head_folder.exists():
            raise FileNotFoundError(
                f"Unable to find microarray template head folder at {self._head_folder!s}"
            )

    def _group_entries(
        self, path
    ) -> OrderedDict[str, OrderedDict[int, list[RawEntry]]]:
        grouped: OrderedDict[str, OrderedDict[int, list[RawEntry]]] = OrderedDict()
        for entry in RawFile(path):
            chromosome = Converter.canonicalize(entry.chromosome)
            if chromosome not in grouped:
                grouped[chromosome] = OrderedDict()
            if entry.position not in grouped[chromosome]:
                grouped[chromosome][entry.position] = []
            grouped[chromosome][entry.position].append(entry)
        return grouped

    def _null_result(self, target: MicroarrayConverterTarget):
        if target in [
            MicroarrayConverterTarget.Ancestry_v1,
            MicroarrayConverterTarget.Ancestry_v2,
        ]:
            return "00"
        return "--"

    def ingest(self, target: MicroarrayConverterTarget):
        header_path = self._head_folder.joinpath(target.name).with_suffix(".txt")
        body_path = self._body_folder.joinpath(target.name).with_suffix(".txt.gz")

        if not header_path.exists():
            raise FileNotFoundError(
                f"Unable to find header file for microarray template at {header_path!s}"
            )
        if not body_path.exists():
            raise FileNotFoundError(
                f"Unable to find body file for microarray template at {body_path!s}"
            )

        template = self._group_entries(body_path)
        lines = []
        for _, chromosome in SequenceOrderer(template.keys()):
            ordered_positions = list(template[chromosome].keys())
            ordered_positions.sort()
            for position in ordered_positions:
                entries = set(x.template_entry for x in template[chromosome][position])
                lines.extend(entries)

        with body_path.with_name(body_path.stem).open("wt") as file:
            file.writelines(lines)

    def analyze(self, target: MicroarrayConverterTarget):
        header_path = self._head_folder.joinpath(target.name).with_suffix(".txt")
        body_path = self._body_folder.joinpath(target.name).with_suffix(".txt")
        duplicate_coords = self._body_folder.joinpath(
            target.name + "_duplicate_coordinates"
        ).with_suffix(".csv")
        non_sequential = self._body_folder.joinpath(
            target.name + "_non_sequential"
        ).with_suffix(".csv")
        same_rsid_different_c = self._body_folder.joinpath(
            target.name + "_same_rsid_different_coordinates"
        ).with_suffix(".csv")
        same_c_different_rsid = self._body_folder.joinpath(
            target.name + "_same_coordinate_different_rsid"
        ).with_suffix(".csv")

        if not header_path.exists():
            raise FileNotFoundError(
                f"Unable to find header file for microarray template at {header_path!s}"
            )
        if not body_path.exists():
            raise FileNotFoundError(
                f"Unable to find body file for microarray template at {body_path!s}"
            )
        template = self._group_entries(body_path)
        lines = []
        lines.append("# Duplicate coordinate report\n")
        lines.append("# Duplicates,Chromosome,Position\n")
        for chromosome, positions in template.items():
            position = 0
            for position, entry in positions.items():
                if len(entry) != 1:
                    lines.append(f"{len(entry)},{chromosome},{position}\n")
        print(
            f"Target {target.name} has a total of {len(lines)-2} duplicate coordinates."
        )

        with duplicate_coords.open("wt") as file:
            file.writelines(lines)

        lines = []
        lines.append("# Non sequential entries\n")
        lines.append("# RSID,Chromosome,Position\n")
        index_by_rsid = dict()
        index_by_coord = dict()
        last_chromosome = None
        last_position = None
        for entry in RawFile(body_path):
            chromosome = Converter.canonicalize(entry.chromosome)
            if chromosome != last_chromosome:
                last_chromosome = chromosome
                last_position = 0

            if entry.id not in index_by_rsid:
                index_by_rsid[entry.id] = []
            index_by_rsid[entry.id].append(entry)
            coord = f"{entry.chromosome},{entry.position}"
            if coord not in index_by_coord:
                index_by_coord[coord] = []
            index_by_coord[coord].append(entry)

            if entry.position < last_position:
                lines.append(f"{entry.id},{entry.chromosome},{entry.position}\n")
            last_position = entry.position

        with non_sequential.open("wt") as file:
            file.writelines(lines)

        print(
            f"Target {target.name} has a total of {len(lines)-2} non sequential entries."
        )

        lines = []
        lines.append("# Same RSID different coordinates\n")
        lines.append("# RSID,Chromosome,Position\n")
        same_rsids: list[list[RawEntry]] = [
            index_by_rsid[x] for x in index_by_rsid if len(index_by_rsid[x]) > 1
        ]

        for same_rsid in same_rsids:
            first_element = same_rsid[0]
            broken_elements = set(
                x
                for x in same_rsid
                if x.chromosome != first_element.chromosome
                or x.position != first_element.position
            )
            if len(broken_elements) > 0:
                lines.append(
                    f"{first_element.id},{first_element.chromosome},{first_element.position}\n"
                )
            for broken_element in broken_elements:
                lines.append(
                    f"{broken_element.id},{broken_element.chromosome},{broken_element.position}\n"
                )

        with same_rsid_different_c.open("wt") as file:
            file.writelines(lines)

        print(
            f"Target {target.name} has a total of {len(lines)-2} entries with same rsID but different coordinates."
        )

        lines = []
        lines.append("# Same coordinates different RSID\n")
        lines.append("# RSID,Chromosome,Position\n")
        same_coords: list[list[RawEntry]] = [
            index_by_coord[x] for x in index_by_coord if len(index_by_coord[x]) > 1
        ]

        for same_coord in same_coords:
            first_element = same_coord[0]
            broken_elements = set(x for x in same_coord if x.id != first_element.id)
            if len(broken_elements) > 0:
                lines.append(
                    f"{first_element.id},{first_element.chromosome},{first_element.position}\n"
                )
            for broken_element in broken_elements:
                lines.append(
                    f"{broken_element.id},{broken_element.chromosome},{broken_element.position}\n"
                )

        with same_c_different_rsid.open("wt") as file:
            file.writelines(lines)

        print(
            f"Target {target.name} has a total of {len(lines)-2} entries with same coordinate but different rsIDs."
        )

    def convert(self, input: Path, target: MicroarrayConverterTarget):
        header_path = self._head_folder.joinpath(target.name).with_suffix(".txt")
        body_path = self._body_folder.joinpath(target.name).with_suffix(".txt.gz")

        if not header_path.exists():
            raise FileNotFoundError(
                f"Unable to find header file for microarray template at {header_path!s}"
            )
        if not body_path.exists():
            raise FileNotFoundError(
                f"Unable to find body file for microarray template at {body_path!s}"
            )

        template = self._group_entries(body_path)
        query_entries = self._group_entries(input)

        lines = []

        # TODO: this is easy to parallelize. Check if it's worth it
        for chromosome in template:
            for position in template[chromosome]:
                template_entries = template[chromosome][position]
                query_entry = None
                if chromosome not in query_entries:
                    query_entry = self._null_result(target)
                elif position not in query_entries[chromosome]:
                    query_entry = self._null_result(target)
                else:
                    if len(query_entries[chromosome][position]) == 0:
                        raise RuntimeError(
                            f"Bug: could not find a query entry for {chromosome} {position}"
                        )
                    query_entry = query_entries[chromosome][position][0].result
                    if len(query_entries[chromosome][position]) != 1:
                        # Sanity check: we've multiple entries for the same chromosome and the same position.
                        # The genotype should be the same. If it's not, warn the user and set a null value.
                        for entry in query_entries[chromosome][position]:
                            if entry.result != query_entry:
                                genomes = set(query_entries[chromosome][position])
                                logging.warn(
                                    f"Found different genotypes for the same position and the same chromosome: {genomes!s}."
                                )
                                query_entry = self._null_result(target)
                                break
                ids = set()
                for template_entry in template_entries:
                    if template_entry.id in ids:
                        # We've an entry with same chromosome/position.
                        # If we need to print it more than once it must at least
                        # have a different ID.
                        continue
                    ids.add(template_entry.id)
                    output_line = TARGET_FORMATTER_MAP[target][0](
                        template_entry.id,
                        template_entry.chromosome,
                        template_entry.position,
                        query_entry,
                    )
                    lines.append(output_line)

        suffix = TARGET_FORMATTER_MAP[target][1].value
        stem = input.stem.replace("_CombinedKit", "")
        output = input.with_name(f"{stem}_{target.name}{suffix}")

        shutil.copy(header_path, output)
        with output.open("a", newline="\n") as f:
            for line in lines:
                f.write(line)


if __name__ == "__main__":
    logging.getLogger().setLevel(logging.DEBUG)
    converter = MicroarrayConverter()

    for value in MicroarrayConverterTarget:
        start = time.time()
        print(f"Ingesting target: {value.name}")
        converter.ingest(value)
        end = time.time()
        print(f"Total: {end-start:.1f}s")



import enum

from PySide6.QtWidgets import QGroupBox, QRadioButton, QVBoxLayout, QWidget


class ExtractTargetSequences(enum.Enum):
    All = enum.auto()
    YOnly = enum.auto()
    MtOnly = enum.auto()
    YAndMT = enum.auto()
    Unmapped = enum.auto()
    Manual = enum.auto()
    Unknown = enum.auto()


class SequenceSelection(QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        self.sequences = QGroupBox(self)
        self.inner_layout = QVBoxLayout(self.sequences)

        self.sequences.setTitle("Select the sequences to extract")

        self.setObjectName(SequenceSelection.__name__)
        self.inner_layout.setObjectName("groupBoxLayout")

        self.sequencesOptions = [
            QRadioButton(label, self.sequences)
            for label in [
                "All sequences",
                "Y only, useful e.g. on yDNA Warehouse, yTree, Cladefinder",
                "MT-Only, useful e.g. on yFull (for biological females), Mitoverse, Haplogrep",
                "Y and MT, useful e.g. on yFull (for biological males)",
                "Unmapped data (microbiome), useful e.g. on Kaiju, CosmosID",
                "Custom, specify the sequences manually",
            ]
        ]

        self.sequencesOptions[0].setObjectName(ExtractTargetSequences.All.name)
        self.sequencesOptions[1].setObjectName(ExtractTargetSequences.YOnly.name)
        self.sequencesOptions[2].setObjectName(ExtractTargetSequences.MtOnly.name)
        self.sequencesOptions[3].setObjectName(ExtractTargetSequences.YAndMT.name)
        self.sequencesOptions[4].setObjectName(ExtractTargetSequences.Unmapped.name)
        self.sequencesOptions[5].setObjectName(ExtractTargetSequences.Manual.name)

        for item in self.sequencesOptions:
            self.inner_layout.addWidget(item)

        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.sequences)
        self.setLayout(self.main_layout)
        self.hide()



import logging
import subprocess
from math import sqrt

from helix.alignment_map.alignment_map_row import AlignmentMapFlag, AlignmentMapRow
from helix.configuration import MANAGER_CFG
from helix.data.alignment_map.alignment_map_file_info import AlignmentMapFileInfo
from helix.data.alignment_stats import AlignmentStats
from helix.data.file_type import FileType
from helix.data.read_type import ReadType
from helix.utility.external import External
from helix.utility.sequencers import Sequencers


class AlignmentStatsCalculator:
    def __init__(
        self,
        path: AlignmentMapFileInfo,
        config=MANAGER_CFG.ALIGNMENT_STATS,
        external: External = External(),
        sequencers: Sequencers = Sequencers(),
        logger=logging.getLogger(__name__),
    ) -> None:
        self.aligned_file = path
        self._config = config
        self._external = external
        self._sequencers = sequencers
        self._logger = logger

    def get_stats(self):
        samples = self._read_samples(self._config.skip, self._config.samples)
        stats = self._process_samples(samples)
        if stats.average_length > 410 and "Nanopore" in stats.sequencer:
            samples.extend(
                self._read_samples(
                    self._config.skip + self._config.samples, self._config.samples * 29
                )
            )
            # Stats takes a fraction of seconds to compute.
            # Re-compute them again
            stats = self._process_samples(samples)
        return stats

    def _read_samples(self, skip, samples_count):
        options = []
        if self.aligned_file.file_type == FileType.CRAM:
            ready_reference = self.aligned_file.reference_genome.ready_reference
            if ready_reference is None:
                raise RuntimeError(
                    "Unable to compute stats because there is no reference available for a CRAM file"
                )
            options.extend(["-T", ready_reference.fasta])
        options.append(self.aligned_file.path)

        process: subprocess.Popen = self._external.samtools(
            ["view", *options], stdout=subprocess.PIPE
        )
        samples = []
        for line in iter(process.stdout.readline, b""):
            if len(samples) == (skip + samples_count):
                break
            samples.append(AlignmentMapRow(line.decode()))
        process.kill()
        # Take the last self.samples if they're enough.
        # Otherwise just take them all.
        start_index = -min(len(samples), samples_count)
        end_index = min(len(samples), skip + samples_count)
        samples = samples[start_index:end_index]
        return samples

    def _process_samples(self, samples: list[AlignmentMapRow]) -> AlignmentStats:
        duplicate_count = 0
        ignored_rows = [
            AlignmentMapFlag.SECONDARY_ALIGNMENT,
            AlignmentMapFlag.NOT_PASSING,
            AlignmentMapFlag.DUPLICATE,
            AlignmentMapFlag.SUPPLEMENTARY_ALIGNMENT,
        ]
        read_type_count = 0
        considered_samples = 0

        if len(samples) == 0:
            self._logger.error("Cannot compute alignment stats as the file is empty")
            return

        # Process the template name for 1st sample to determine the sequencer
        # as it should not give a different result on the other samples.
        sequencer = self._sequencers.determine_sequencer(samples[0].query_template_name)

        # Used to compute stats for read length
        count_length = 0
        average_length = 0
        squared_deviation_length = 0

        # Used to compute stats for insert size
        count_insert_size = 0
        average_insert_size = 0
        squared_deviation_insert_size = 0

        # Used to compute stats for alignment quality
        count_quality = 0
        average_quality = 0
        squared_deviation_quality = 0

        for sample in samples:
            if sample.flag & AlignmentMapFlag.DUPLICATE:
                duplicate_count += 1
            if any(sample.flag & x for x in ignored_rows):
                continue

            if sample.flag & AlignmentMapFlag.MULTIPLE_SEGMENTS:
                read_type_count += 1
            else:
                read_type_count += -1

            considered_samples += 1

            # Compute stats for read length
            read_length = len(sample.sequence)
            # The sequence can be '*' sometimes. Ignore in that case.
            # Ref.: Page 9 of the standard.
            if read_length > 1:
                count_length += 1
                delta_length = read_length - average_length
                average_length += delta_length / count_length
                new_delta_length = read_length - average_length
                squared_deviation_length += delta_length * new_delta_length

            # Compute stats for insertion size
            if sample.mate_sequence_name == "=":
                template_length = sample.template_length
                if 0 < template_length < 50000:
                    count_insert_size += 1
                    delta_insert_size = template_length - average_insert_size
                    average_insert_size += delta_insert_size / count_insert_size
                    new_delta_insert_size = template_length - average_insert_size
                    squared_deviation_insert_size += (
                        delta_insert_size * new_delta_insert_size
                    )

            # Compute stats for alignment quality
            count_quality += 1
            delta_quality = sample.mapping_quality - average_quality
            average_quality += delta_quality / count_quality
            new_delta_quality = sample.mapping_quality - average_quality
            squared_deviation_quality += delta_quality * new_delta_quality

        # Establish the read type based on the majority of samples
        majority = considered_samples * 0.5
        read_type = ReadType.Unknown
        if read_type_count > majority:
            read_type = ReadType.Paired
        elif read_type_count < -majority:
            read_type = ReadType.Single

        if count_length <= 2:
            self._logger.error(
                "Unable to compute read length stats as the number of valid samples is less than 2."
            )
            return None
        if count_insert_size <= 2 and read_type == ReadType.Paired:
            self._logger.error(
                "Unable to compute insert size stats as the number of valid samples is less than 2."
            )
            return None
        if count_quality <= 2:
            self._logger.error(
                "Unable to compute alignment quality stats as the number of valid samples is less than 2."
            )
            return None
        if read_type == ReadType.Unknown:
            self._logger.error(
                "Unable to determine read type as there's a similar number of single vs paired reads."
            )
            return None

        stats = AlignmentStats()
        stats.samples_count = self._config.samples
        stats.skipped_samples = self._config.skip
        stats.read_type = read_type
        stats.duplicate = duplicate_count
        stats.sequencer = sequencer

        if stats.read_type == ReadType.Paired:
            stats.count_insert_size = count_insert_size
            stats.average_insert_size = average_insert_size
            stats.standard_dev_insert_size = sqrt(
                squared_deviation_insert_size / (count_insert_size - 1)
            )
        else:
            stats.count_insert_size = 0
            stats.average_insert_size = 0
            stats.standard_dev_insert_size = 0

        stats.count_length = count_length
        stats.average_length = average_length
        stats.standard_dev_length = sqrt(squared_deviation_length / (count_length - 1))

        stats.count_quality = count_quality
        stats.average_quality = average_quality
        stats.standard_dev_quality = sqrt(
            squared_deviation_quality / (count_quality - 1)
        )
        return stats



import logging
import os
import shlex
import shutil
import subprocess
import sys
from pathlib import Path

from helix.configuration import MANAGER_CFG
from helix.progress.process_io_monitor import ProcessIOMonitor

logger = logging.getLogger(__name__)

if sys.platform == "win32":
    third_party = str(MANAGER_CFG.EXTERNAL.root)
    if third_party not in os.environ["PATH"]:
        os.environ["PATH"] += ";" + third_party
    if ".JAR" not in os.environ["PATHEXT"]:
        os.environ["PATHEXT"] += ";.JAR"


def exe(f, interpreter=[]):
    """This decorator will return a function that will try to launch
    an executable from disk that has the same name of the function
    it's decorating, passing the arguments that were received when
    the function was invoked.

    Args:
        f (Callable): function to decorate.
    """

    def execute_binary(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        args = [*interpreter, shutil.which(f.__name__), *[str(x) for x in args]]

        if wait:
            # Force stdout/stderr to be PIPE as we
            # need to collect the output
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE
        logger.debug(f"Calling: {shlex.join(args)}")

        # Force windows to hide the prompt window
        startup_info = None
        if sys.platform == "win32":
            startup_info = subprocess.STARTUPINFO()
            startup_info.dwFlags |= subprocess.STARTF_USESHOWWINDOW

        output = subprocess.Popen(
            args,
            stdout=stdout,
            stdin=stdin,
            stderr=stderr,
            startupinfo=startup_info,
            text=text,
        )
        if io is not None:
            monitor = ProcessIOMonitor(output, io)
            monitor.start()
        if wait is True:
            out, err = output.communicate()
            if output.returncode != 0:
                raise RuntimeError(f"Call to {f.__name__} failed: {err.decode()}")
            return out
        return output

    decorated = execute_binary
    decorated.wrapper = exe
    decorated.__name__ = f.__name__
    return decorated


def jar(f):
    """Same thing as run but this deals with .jar files
    automatically, invoking java from PATH and specifying
    the right arguments, including the full path of the .jar
    file.

    Args:
        f (callable): function to decorate

    Returns:
        callable: Decorated function
    """
    full_path = shutil.which(f.__name__)
    if full_path is None:
        # Should raise NotImplementedError
        f.wrapper = jar
        return f
    full_path = Path(".", full_path)
    full_path = full_path.with_suffix(".jar")
    f.__name__ = str(full_path)
    decorated = exe(f, ["java", "-jar"])
    decorated.wrapper = jar
    return decorated


class External:
    """Wrapper around 3rd party executables

    TODO: make this class contains only wrappers around exe/jar files and
    move the rest of the logic somewhere else (e.g., Samtools class with
    view(), consensus(), ..., a Haplogrep class, a gzip class etc.).
    """

    def __init__(self, config=MANAGER_CFG.EXTERNAL) -> None:
        self._config = config
        if not self._config.root.exists():
            raise FileNotFoundError(
                f"Unable to find root directory for External: {str(self._config.root)}"
            )
        if str(self._config.root) not in os.environ["PATH"]:
            os.environ["PATH"] += ";" + str(self._config.root)

        self._htsfile = "htsfile"

    def get_file_type(self, path: Path):
        process = subprocess.run([self._htsfile, path], capture_output=True, check=True)
        return process.stdout.decode("utf-8")

    def haplogrep_classify(self, vcf_file, output_file):
        output = self.haplogrep(
            ["classify", "--in", vcf_file, "--format", "vcf", "--out", output_file],
            wait=True,
        )
        output.decode("utf-8")
        return output

    # Starting from here all the functions are
    # just calling executables with the same name.
    # See the implementation of @exe and @jar decorator
    # for more details.

    @exe
    def gzip(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def bgzip(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def samtools(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def bwa(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def bwamem2(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def minimap2(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def fastp(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def bcftools(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @exe
    def tabix(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @jar
    def haplogrep(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @jar
    def FastQC(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @jar
    def picard(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()

    @jar
    def DISCVRSeq(
        self,
        args=[],
        stdout=None,
        stdin=None,
        stderr=None,
        wait=False,
        io=None,
        text=False,
    ):
        raise FileNotFoundError()



from enum import IntEnum


class DepthBin(IntEnum):
    Zero = 0
    Between0And3 = 1
    Between3And7 = 2
    MoreThan7 = 3


class CoverageStats:
    def __init__(self) -> None:
        self.sequence_name: str = None
        self.bin_entries: dict[DepthBin, int] = None
        self.bin_sum: dict[DepthBin, int] = None
        self.non_zero_percentage: float = 0
        self.non_zero_average: float = 0
        self.all_average: float = 0



# pkgutil shenanigans to make python load helix.third_party from site-packages
#
# tl;dr helix.third_party contains some executables for windows and is installed
# from another pypi package (doublehelix-external).
# When importing helix.third_party from any files inside Helix, python will expect
# to find helix.third_party INSIDE this package and not in site-packages. Code
# below will make this package a "Native Namespace Package" (PEP 420) and will
# make it possible to have different packages contributing to the same namespace
# (helix). Only downside is the the IDE won't recognize helix.third_party, which is
# not a big deal as it doesn't contain any python code.

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)



from pydantic import BaseModel


class AlignmentMapHeaderProgram(BaseModel):
    id: str = None
    name: str = None
    command_line: str = None
    previous: "AlignmentMapHeaderProgram" = None
    description: str = None
    program_version: str = None



import subprocess
import sys
from importlib.metadata import version


class Updater:

    def __init__(self) -> None:
        self.packages = ["Helix", "doublehelix-external"]
        # Package Name -> (Current v., Latest v.)
        self.versions = {x: (version(x), self.check(x)) for x in self.packages}

    def check(self, package_name):
        latest_version = subprocess.check_output(
            [sys.executable, "-m", "pip", "index", "versions", package_name]
        ).decode("utf-8")
        latest_version = latest_version.split("Version: ")[1].split("\n")[0]
        return latest_version

    def update(self, package_name, versions):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "--upgrade", package_name]
        )
        print(f"Upgraded {package_name} from {versions[0]} to {versions[1]}")

        sys.exit()


if __name__ == "__main__":
    a = Updater()
    pass



from collections import OrderedDict

from helix.data.genome import Genome
from helix.data.sequence import Sequence
from helix.fasta.reference import Reference, ReferenceStatus


def test_query_is_a_perfect_match():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 123, "a", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, None), [sequences_1[0]]),
        ]
    )

    assert Reference(query_sequences).status == ReferenceStatus.Downloadable


def test_query_is_superset():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 123, "a", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, None), [sequences_1[0]]),
            (Sequence("chr2", 124, None), []),
        ]
    )

    assert Reference(query_sequences).status == ReferenceStatus.Unknown


def test_genome_is_superset():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 123, "a", genome_1),
        Sequence("chr2", 124, "b", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, None), [sequences_1[0]]),
        ]
    )

    ref = Reference(query_sequences)
    assert ref.status == ReferenceStatus.Buildable
    assert len(ref.matching) == 0


def test_genome_has_different_md5():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 123, "a", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, "b"), [sequences_1[0]]),
        ]
    )

    assert len(Reference(query_sequences).matching) == 0


def test_genome_has_different_length():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 122, "a", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, None), [sequences_1[0]]),
        ]
    )

    assert len(Reference(query_sequences).matching) == 0


def test_only_one_genome_matches():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 122, "a", genome_1),
        Sequence("chr2", 123, "b", genome_1),
    ]
    genome_1.sequences = sequences_1

    genome_2 = Genome("https://reference_2.com/fasta.fa", source="Fake", build="38")
    sequences_2 = [
        Sequence("chr1", 122, "a", genome_2),
    ]
    genome_2.sequences = sequences_2

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 122, "a"), [sequences_1[0], sequences_2[0]]),
            (Sequence("chr2", 123, "b"), [sequences_1[1]]),
        ]
    )

    assert len(Reference(query_sequences).matching) == 1


def test_genome_has_different_length_and_md5():
    genome_1 = Genome("https://reference_1.com/fasta.fa", source="Fake", build="38")
    sequences_1 = [
        Sequence("chr1", 123, "a", genome_1),
    ]

    genome_1.sequences = sequences_1

    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 122, "b"), [sequences_1[0]]),
        ]
    )

    assert len(Reference(query_sequences).matching) == 0


def test_no_matches():
    query_sequences = OrderedDict(
        [
            (Sequence("chr1", 123, None), []),
        ]
    )

    assert Reference(query_sequences).status == ReferenceStatus.Unknown



import logging
import os
import sys
import webbrowser
from pathlib import Path

from PySide6.QtCore import QModelIndex, Signal
from PySide6.QtWidgets import (
    QApplication,
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QTableWidgetItem,
    QAbstractScrollArea,
)

from helix.adapters.alignment_stats_adapter import AlignmentStatsAdapter
from helix.adapters.coverage_stats_adapter import CoverageStatsAdapter
from helix.adapters.header_adapter import (
    HeaderCommentsAdapter,
    HeaderProgramsAdapter,
    HeaderReadGroupAdapter,
    HeaderSequenceAdapter,
)
from helix.adapters.index_stats_adapter import IndexStatsAdapter
from helix.adapters.reference_adapter import ReferenceAdapter
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.alignment_map.depth_analyzer import CoverageStatsCalculator
from helix.alignment_map.index_stats_calculator import IndexStatsCalculator
from helix.configuration import MANAGER_CFG
from helix.data.sorting import Sorting
from helix.fasta.reference import ReferenceStatus
from helix.gui.extract.extract_wizard import ExtractWizard
from helix.gui.table_dialog import ListTableDialog, TableDialog
from helix.gui.ui_form import Ui_MainWindow
from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.reference.repository_manager import Repository
from helix.utility.external import External
from helix.utility.samtools import Samtools
from helix.utility.shortcut import Shortcut
from helix.progress.simple_worker import SimpleWorker


class HelixWindow(QMainWindow):

    _percentage_updated = Signal(int)
    _message_updated = Signal(str)
    _operation_ended = Signal()
    _coverage_ready = Signal()

    def launch():
        app = QApplication(sys.argv)
        widget = HelixWindow()
        widget.show()
        sys.exit(app.exec())

    def __init__(
        self,
        parent=None,
        config=MANAGER_CFG.GENERAL,
        external=External(),
        samtools=Samtools(),
        repository=Repository(),
    ):
        super().__init__(parent)
        self._external = external
        self._samtools = samtools
        self._repository = repository
        self.config = config
        self.current_file = None
        self._logger = logging.getLogger("main")

        self.switch_to_main()
        self.ui.actionDocumentation.triggered.connect(self.on_doc)
        self.ui.actionOpen.triggered.connect(self.on_open)
        self.ui.actionExit.triggered.connect(self.on_close)
        self.ui.actionSettings.triggered.connect(self.on_settings)
        self.ui.actionCreate_launcher_on_desktop.triggered.connect(self.create_launcher)

    def create_launcher(self):
        path = Shortcut().create()
        message_box = QMessageBox()
        message_box.setWindowTitle("Created")
        message_box.setText(f"A link to Helix was created at {path}")
        message_box.setStandardButtons(QMessageBox.StandardButton.Ok)
        message_box.exec()

    def switch_to_main(self):
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        self.ui.fileInformationTable.doubleClicked.connect(self.file_item_clicked)
        self.ui.extract.clicked.connect(self.export)
        self.ui.stop.clicked.connect(self.stop)

        self._message_updated.connect(self.ui.statusbar.showMessage)
        self._percentage_updated.connect(self.ui.progress.setValue)
        self._coverage_ready.connect(self._show_coverage_stats)
        self._operation_ended.connect(self._prepare_ready)

        self._long_operation: SimpleWorker = None
        self._prepare_ready(self.config.last_path)

    def stop(self):
        if self._long_operation is not None:
            self._long_operation.kill()
            self._long_operation = None
            self._worker = None
        self._prepare_ready()

    def export(self):
        if self.current_file is None:
            return
        self._prepare_long_operation("Prepare exporting")
        dialog = ExtractWizard(self.current_file, self, progress=self._set_progress)
        dialog.exec()
        if dialog._worker is not None:
            self._long_operation = dialog
            self._worker = dialog._worker
        else:
            self._prepare_ready()

    def closeEvent(self, event=False):
        MANAGER_CFG.save()
        event.accept()

    def on_close(self):
        MANAGER_CFG.save()
        self.close()

    def on_doc(self):
        webbrowser.open("https://helix.readthedocs.io", 2)

    def on_open(self, file: Path = False):
        if not file:
            title = "Open file"
            type = "Aligned File (*.bam *.sam *.cram);;"
            type += "Unaligned file (*.fastq);;"
            type += "References (*.fa *.gz *.fna *.fasta)"
            last_path = str(Path.home())
            if self.config.last_path is not None:
                if self.config.last_path.is_dir():
                    last_path = str(self.config.last_path)
                else:
                    last_path = str(Path(self.config.last_path).parent)
            file = QFileDialog.getOpenFileName(self, title, last_path, type)
            if file[0] == "":
                return
            file = Path(file[0])

        if not file.exists():
            return
        if not file.is_file():
            return

        self.config.last_path = file
        self.load_aligned(file)

    def on_settings(self):
        pass

    def load_aligned(self, file):
        self.current_file = AlignmentMapFile(Path(file))
        info = self.current_file.file_info

        sorted_string = info.sorted.name
        if info.sorted in [Sorting.Unknown, Sorting.Unsorted]:
            sorted_string += " (Click to sort)"

        indexed_string = str(info.indexed)
        gender_string = info.gender.name
        if not info.indexed:
            indexed_string += " (Click to index)"
            gender_string += " (File is not indexed)"
        click_to_open = "(Click here to open)"

        reference_str = (
            f"{info.reference_genome.build}, {info.reference_genome.status.name} "
        )
        reference_str += "(Click for details)"

        size = self.current_file.path.stat().st_size
        size /= 1024**3
        size = f"{size:.1f} GB"

        rows = [
            ("Directory", str(self.current_file.path.parent)),
            ("Filename", str(self.current_file.path.name)),
            ("Size", size),
            ("File Type", info.file_type.name),
            ("Reference", reference_str),
            ("Gender", gender_string),
            ("Sorted", sorted_string),
            ("Indexed", indexed_string),
            ("Mitochondrial DNA Model", info.mitochondrial_dna_model.name),
            ("Header", click_to_open),
            ("Alignment stats", click_to_open),
            ("Index stats", click_to_open),
            ("Coverage stats", click_to_open),
        ]

        self.ui.fileInformationTable.setRowCount(len(rows))
        self.ui.fileInformationTable.setColumnCount(1)
        for index, row in enumerate(rows):
            header, value = row
            self.ui.fileInformationTable.setVerticalHeaderItem(
                index, QTableWidgetItem(header)
            )
            self.ui.fileInformationTable.setItem(index, 0, QTableWidgetItem(value))

    def file_item_clicked(self, index: QModelIndex):
        if self.current_file is None:
            return
        handlers = {
            "Directory": self._open_dir,
            "Sorted": self._do_sorting,
            "Indexed": self._do_indexing,
            "Reference": self._show_reference,
            "Header": self._show_header,
            "Alignment stats": self._show_alignment_stats,
            "Index stats": self._show_index_stats,
            "Coverage stats": self._show_coverage_stats,
        }
        item = self.ui.fileInformationTable.verticalHeaderItem(index.row())
        if item.text() not in handlers:
            return
        handlers[item.text()]()

    def _open_dir(self):
        if getattr(os, "startfile", None):
            os.startfile(self.current_file.file_info.path.parent)

    def _do_sorting(self):
        if self.current_file is None:
            return
        if self.current_file.file_info.sorted not in [
            Sorting.Unknown,
            Sorting.Unsorted,
        ]:
            return

        choice = self._yn_message_box(
            "File is not sorted",
            "Do you want to sort the file?",
            "This may take a while.",
        )
        if choice == QMessageBox.StandardButton.No:
            return

    def _do_indexing(self, user_informed=False):
        if self.current_file is None:
            return
        if self.current_file.file_info.indexed:
            return
        if not user_informed:
            choice = self._yn_message_box(
                "File is not indexed",
                "Do you want to index the file?",
                "This may take a while.",
            )
            if choice == QMessageBox.StandardButton.No:
                return
        self._prepare_long_operation("Start indexing.")
        progress = BaseProgressCalculator(
            self._set_progress, self.current_file.path.stat().st_size, "Indexing"
        )

        try:
            self._long_operation = SimpleWorker(
                self._samtools.index,
                self.current_file.path,
                io=progress.compute_on_read_bytes,
            )
        except Exception as e:
            self._logger.error(f"Error while initializing indexing: {e!s}")
            self._prepare_ready()

    def _set_progress(self, label, percentage):
        # Memento: this is always called from another thread.
        #   Don't put anything here that is updating the UI if
        #   you don't want the app to crash spectacularly and
        #   you don't want to spend the next 2 hours troubleshooting
        #   the issue. Manage every interaction with the UI with
        #   signal/slots.
        if label is None:
            self._operation_ended.emit()
            return
        self._percentage_updated.emit(percentage)
        self._message_updated.emit(label)

    def _prepare_long_operation(self, message):
        self.ui.progress.setMaximum(100)
        self.ui.progress.setMinimum(0)
        self.ui.progress.setValue(0)
        self.ui.progress.show()
        self.ui.stop.setEnabled(True)
        self.ui.stop.show()
        self.ui.extract.hide()
        self.ui.statusbar.showMessage(message)

    def _prepare_ready(self, path=None):
        self._long_operation = None
        self.ui.progress.hide()
        self.ui.stop.setEnabled(False)
        self.ui.stop.hide()
        if self.current_file is not None:
            self.on_open(self.current_file.path)
        elif path is not None:
            self.on_open(path)
        self.ui.statusbar.showMessage("Ready")
        self.ui.extract.show()

    def _yn_message_box(self, title, text, info=None):
        prompt = QMessageBox()
        prompt.setWindowTitle(title)
        prompt.setText(text)
        if info is not None:
            prompt.setInformativeText(info)

        buttons = QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
        prompt.setStandardButtons(buttons)
        prompt.setDefaultButton(QMessageBox.StandardButton.Yes)
        return QMessageBox.StandardButton(prompt.exec())

    def _show_index_stats(self):
        if self.current_file is None:
            return

        if not self.current_file.file_info.indexed:
            choice = self._yn_message_box(
                "File is not indexed",
                "Index stats are not available if the file is not indexed.",
                "Do you want to index the file? This may take a while.",
            )
            if choice == QMessageBox.StandardButton.No:
                return
            self._do_indexing(user_informed=True)

        if self.current_file.file_info.index_stats is None:
            choice = self._yn_message_box(
                "Index stats not available",
                "Index stats needs to be calculated for this file.",
                "Do you want to calculate the stats? This may take a while.",
            )
            if choice == QMessageBox.StandardButton.No:
                return

            indexed_stats = IndexStatsCalculator(self.current_file.path)
            self.current_file.file_info.index_stats = indexed_stats.get_stats()
            self.current_file.file_info.gender = self.current_file.get_gender(
                self.current_file.file_info.index_stats
            )
            self.current_file.save_meta()

        index_stats = self.current_file.file_info.index_stats
        dialog = TableDialog("Index Statistics", self)
        dialog.set_data(IndexStatsAdapter.adapt(index_stats))
        dialog.exec()

    def _show_coverage_stats(self):
        if self.current_file is None:
            return

        if self.current_file.file_info.coverage_stats is None:
            choice = self._yn_message_box(
                "Coverage stats are not computed",
                "Do you want to calculate coverage stats? This may take a while.",
            )
            if choice == QMessageBox.StandardButton.No:
                return
            self._prepare_long_operation("Preparing to compute coverage stats")
            self._worker = SimpleWorker(self._compute_coverage_stats)
        else:
            coverage_statistics = self.current_file.file_info.coverage_stats
            dialog = TableDialog("Coverage Statistics", self)
            dialog.tableWidget.setSizeAdjustPolicy(
                QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents
            )
            dialog.set_data(CoverageStatsAdapter.adapt(coverage_statistics))
            dialog.tableWidget.resizeColumnsToContents()
            dialog.exec()

    def _compute_coverage_stats(self):
        self._long_operation = CoverageStatsCalculator(progress=self._set_progress)
        self.current_file.file_info.coverage_stats = self._long_operation.get_stats(
            self.current_file
        )
        self.current_file.save_meta()
        self._coverage_ready.emit()

    def _show_reference(self):
        if self.current_file is None:
            return
        reference = self.current_file.file_info.reference_genome
        if (
            reference.ready_reference is None
            and reference.status == ReferenceStatus.Downloadable
        ):
            choice = self._yn_message_box(
                "Reference is not on disk",
                "A matching reference genome for this aligned file is available to be downloaded",
                "Would you like to download it?",
            )
            if choice == QMessageBox.StandardButton.Yes:
                self._perform_reference_download()
        reference = self.current_file.file_info.reference_genome
        dialog = TableDialog("Reference genome", self)
        dialog.tableWidget.setSizeAdjustPolicy(
            QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents
        )
        dialog.set_data(ReferenceAdapter.adapt(reference))
        dialog.tableWidget.resizeColumnsToContents()
        dialog.exec()

    def _perform_reference_download(self):
        if self.current_file is None:
            return
        status = self.current_file.file_info.reference_genome.status
        if status != ReferenceStatus.Downloadable:
            return
        reference = self.current_file.file_info.reference_genome
        for match in reference.matching:
            try:
                SimpleWorker(self._repository.download, match, self._set_progress)
                break
            except Exception as e:
                self._logger.error(f"Error while downloading the reference: {e!s}")

    def _show_alignment_stats(self):
        if self.current_file is None:
            return

        if self.current_file.file_info.alignment_stats is None:
            self._yn_message_box(
                "Alignment stats not available",
                "Cannot compute alignment stats if there are no references available.",
                "",
            )
            return
        alignment_stats = self.current_file.file_info.alignment_stats
        dialog = TableDialog("Alignment statistics", self)
        dialog.set_data(AlignmentStatsAdapter.adapt(alignment_stats))
        dialog.exec()

    def _show_header(self):
        if self.current_file is None:
            return

        sequences = self.current_file.header.sequences.values()
        programs = self.current_file.header.programs
        read_groups = self.current_file.header.read_groups
        comments = self.current_file.header.comments

        dialog = ListTableDialog("Header", self)
        dialog.set_data(
            {
                "Sequences": HeaderSequenceAdapter.adapt(sequences),
                "Programs": HeaderProgramsAdapter.adapt(programs),
                "Read groups": HeaderReadGroupAdapter.adapt(read_groups),
                "Comments": HeaderCommentsAdapter.adapt(comments),
            }
        )
        dialog.exec()



import enum
import webbrowser

from PySide6.QtCore import QMetaObject, Qt
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
)

from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.alignment_map.variant_caller import VariantCaller
from helix.data.file_type import FileType
from helix.gui.extract.format_selection import ExtractTargetFormat, FormatSelection
from helix.gui.extract.microarray_selection import MicroarraySelection
from helix.gui.extract.sequence_selection import (
    ExtractTargetSequences,
    SequenceSelection,
)
from helix.progress.simple_worker import SimpleWorker
from helix.renderers.html_aligned_file_report import HTMLAlignedFileReport


class ExtractTargetSteps(enum.Enum):
    FormatSelection = enum.auto()
    SequenceSelection = enum.auto()
    MicroarraySelection = enum.auto()


class ExtractWizard(QDialog):
    def __init__(self, current_file, parent=None, progress=None) -> None:
        self.current_file: AlignmentMapFile = current_file
        super().__init__(parent, Qt.WindowType.Dialog)
        self._progress = progress

        self.resize(648, 600)
        self.setObjectName("extract")
        self.setWindowTitle("Extract data")

        # [Back] <--> [Next]
        self.back_button = QPushButton("Back", self)
        self.back_button.setObjectName("backButton")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        self.next_button = QPushButton("Next", self)
        self.next_button.setObjectName("nextButton")

        self._format_selection = FormatSelection(
            self, self.current_file.file_info.file_type
        )
        self._sequence_selection = SequenceSelection(self)
        self._microarray_selection = MicroarraySelection(self)

        self.mainLayout = QGridLayout(self)
        self.mainLayout.setObjectName("mainLayout")

        # 1st row
        self.main = self._format_selection
        self.mainLayout.addWidget(self.main, 0, 0, 1, 3)
        self.main.show()

        # 2nd row [BACK] (space) [NEXT]
        self.mainLayout.addWidget(self.back_button, 1, 0, 1, 1)
        self.mainLayout.addItem(self.horizontalSpacer, 1, 1, 1, 1)
        self.mainLayout.addWidget(self.next_button, 1, 2, 1, 1)

        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        QMetaObject.connectSlotsByName(self)
        self.next_button.clicked.connect(self.next)
        self.back_button.clicked.connect(self.back)

        self._format_handlers = {
            ExtractTargetFormat.Microarray: self._to_microarray,
            ExtractTargetFormat.BAM: self._to_bam,
            ExtractTargetFormat.SAM: self._to_sam,
            ExtractTargetFormat.CRAM: self._to_cram,
            ExtractTargetFormat.FASTQ: self._to_fastq,
            ExtractTargetFormat.FASTA: self._to_fasta,
            ExtractTargetFormat.HTML: self._to_html,
            ExtractTargetFormat.VCF: self._to_vcf,
            ExtractTargetFormat.Unknown: self._to_unknown,
        }

        # TODO: move to test
        assert all(x in self._format_handlers for x in ExtractTargetFormat)

    def _to_bam(self):
        if self.current_file is None:
            return

        self._worker = SimpleWorker(
            self.current_file.convert, FileType.BAM, progress=self._progress
        )
        self.close()

    def _to_sam(self):
        if self.current_file is None:
            return
        self._worker = SimpleWorker(
            self.current_file.convert, FileType.SAM, progress=self._progress
        )
        self.close()

    def _to_cram(self):
        if self.current_file is None:
            return

        self._worker = SimpleWorker(
            self.current_file.convert, FileType.CRAM, progress=self._progress
        )
        self.close()

    def _to_fasta(self):
        if self.current_file is None:
            return

        self._worker = SimpleWorker(
            self.current_file.convert, FileType.FASTA, progress=self._progress
        )
        self.close()

    def _to_fastq(self):
        if self.current_file is None:
            return
        self.close()

    def _to_html(self):
        if self.current_file is None:
            return

        html_page = HTMLAlignedFileReport.from_aligned_file(self.current_file)
        current_path = self.current_file.path
        extensions = "".join(current_path.suffixes[:-1])
        name = current_path.stem + extensions + ".html"
        target = current_path.with_name(name)

        with target.open("wt") as f:
            f.write(html_page)
        self._progress(None, None)
        webbrowser.open(target)
        self.close()

    def _to_microarray(self):
        pass

    def _to_vcf(self):
        if self.current_file is None:
            return

        long_operation = VariantCaller(progress=self._progress)
        self._worker = SimpleWorker(
            long_operation.call,
            self.current_file,
        )
        self.close()

    def _to_unknown(self):
        raise RuntimeError("BUG: Invalid target format")

    def move_to(self, target):
        if self.main is not None:
            self.main.hide()
        self.mainLayout.replaceWidget(self.main, target)
        self.main = target
        self.main.show()

    def _format_selected(self):
        selected = [x.isChecked() for x in self._format_selection._format_options]
        if not any(selected):
            return
        assert len([x for x in selected if x]) == 1
        index = selected.index(True)
        button = self._format_selection._format_options[index]
        self._target_format = ExtractTargetFormat[button.objectName()]

        if self._target_format == ExtractTargetFormat.HTML:
            self._to_html()
        elif self._target_format == ExtractTargetFormat.Microarray:
            self.move_to(self._microarray_selection)
        else:
            self.move_to(self._sequence_selection)

    def _sequence_selected(self):
        selected = [x.isChecked() for x in self._sequence_selection.sequencesOptions]
        if not any(selected):
            return
        assert len([x for x in selected if x]) == 1
        index = selected.index(True)
        button = self._sequence_selection.sequencesOptions[index]
        self._target_sequences = ExtractTargetSequences[button.objectName()]
        self._format_handlers[self._target_format]()

    def back(self):
        if self.main.objectName() == FormatSelection.__name__:
            pass
        elif self.main.objectName() == SequenceSelection.__name__:
            self.move_to(self._format_selection)
        elif self.main.objectName() == MicroarraySelection.__name__:
            self.move_to(self._format_selection)

    def next(self):
        if self.main.objectName() == FormatSelection.__name__:
            self._format_selected()
        elif self.main.objectName() == SequenceSelection.__name__:
            self._sequence_selected()
        elif self.main.objectName() == MicroarraySelection.__name__:
            pass



import random
import textwrap
from pathlib import Path

from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.reference.genome_metadata_loader import MetadataLoader


class FakeReferenceGenomeGenerator:

    ACGT_PERCENTAGE = {"A": 0.309, "T": 0.294, "G": 0.199, "C": 0.198}

    def __init__(self, fasta_dictionary: Path, read_length: int):
        self.read_length = read_length
        self.dictionary = AlignmentMapHeader.load_from_file(fasta_dictionary)

    def generate_fasta(self, length, filename: Path = Path("fake_genome.fasta")):
        bases = list(FakeReferenceGenomeGenerator.ACGT_PERCENTAGE.keys())
        bases_weights = list(FakeReferenceGenomeGenerator.ACGT_PERCENTAGE.values())

        total_bases = sum([x.length for x in self.dictionary.entries.values()])

        with open(filename, "w") as f:
            for name, entry in self.dictionary.entries.items():
                length_adapted = int((entry.length / total_bases) * length)
                reminder = length_adapted % self.read_length
                length_adapted += self.read_length - reminder
                sequence = "".join(
                    random.choices(bases, weights=bases_weights, k=length_adapted)
                )
                lines = "\n".join(textwrap.wrap(sequence, 60))
                if len(sequence) == 0:
                    print(f"Got 0bp to generate for sequence {name}. Skipping")
                    continue
                print(f"Got {length_adapted}bp for sequence {name}")
                f.write(f">{name}\n")
                f.write(lines + "\n")


# Example usage:
if __name__ == "__main__":
    metadata = MetadataLoader()

    generator = FakeReferenceGenomeGenerator(Path("reference_genome.fa.dict"), 150)
    generator.generate_fasta(100000)



import typing

from helix.fasta.letter_run import LetterRun


class LetterRunCollection:
    """Represent a sequence containing runs of a specific letter."""

    def __init__(self, name: str, length: int) -> None:
        if length <= 0:
            raise IndexError("Length should be greater than zero.")
        self.runs: typing.List[LetterRun] = []
        self.name: str = name
        self.length: int = length
        self._current_run: LetterRun = None
        self._is_ended: bool = False

    def filter(self, criteria: typing.Callable[[LetterRun], bool]):
        return [x for x in self.runs if criteria(x)]

    def open_run(self, position: int) -> None:
        if self._is_ended:
            raise RuntimeError("Trying to open a run on an ended sequence.")
        if self._current_run is not None:
            raise RuntimeError("Trying to open a an already opened run of Ns.")
        if position < 0:
            raise IndexError("Trying to open a run with a negative position.")

        if len(self.runs) > 0:
            previous_end = self.runs[-1].start + self.runs[-1].length
            if position < previous_end + 1:
                raise ValueError("Trying to open a run overlapping with previous runs.")

        self._current_run = LetterRun()
        self._current_run.open(position)

    def is_run_open(self) -> bool:
        return self._current_run is not None

    def close_run(self, position: int) -> None:
        if self._current_run is None:
            raise RuntimeError("Trying to close an already closed run.")
        if position <= self._current_run.start:
            raise RuntimeError(
                f"Expected a position greater than start for closing a run: {position}<={self._current_run.start}"
            )
        if position > self.length:
            raise RuntimeError(
                f"Trying to close a run with a position {position}, that is greater than the whole length of the sequence {self.length}."
            )
        run_length = position - self._current_run.start
        self._current_run.close(run_length)
        self.runs.append(self._current_run)
        self._current_run = None

    def end(self, position: int) -> None:
        if self.is_run_open():
            self.close_run(position)
        self._is_ended = True

        if position != self.length:
            raise ValueError(
                f"Expected {self.length} base pairs in this sequence but processed {position}."
            )



import enum
from pathlib import Path

from helix.utility.external import External


class GZipAction(enum.Enum):
    Compress = 0
    Decompress = 1


class GZip:
    def __init__(self, external=External()) -> None:
        self._external = external

    def _gzip_filename(self, input: Path, action: GZipAction):
        if action == GZipAction.Compress:
            return Path(str(input) + ".gz")
        elif action == GZipAction.Decompress:
            if len(input.suffixes) == 0:
                raise RuntimeError(
                    f"Unable to determine decompressed filename, invalid filename {str(input)} (no extensions)."
                )
            ext = "".join(input.suffixes[:-1])
            return input.with_name(input.stem + ext)
        else:
            raise RuntimeError(f"Action {action.name} not supported.")

    def gzip(
        self,
        input: Path,
        output: Path,
        action: GZipAction = GZipAction.Decompress,
    ) -> Path:
        # TODO: move this logic somewhere else.
        if output.exists():
            raise RuntimeError(
                f"Trying to decompress {str(input)} but the destination file {str(output)} exists."
            )
        inferred_filename = self._gzip_filename(input, action)

        action_flags = {GZipAction.Compress: "", GZipAction.Decompress: "-d"}

        process = self._external.gzip([action_flags[action], str(input)])
        process.wait()

        # RAFZ format is libz compatible but will make gzip sometime exit
        # with a != 0 code, complaining about "trailing garbage data".
        # This is not a real error, as the file is decompressed anyway.
        # The issue is potentially fixable by truncating the file, but
        # there's no practical advantage in doing so. If we fall in this
        # situation, ignore the error.
        if process.returncode != 0:
            if "trailing garbage" not in process.stderr.decode():
                raise RuntimeError(
                    f"gzip exited with return code {process.returncode}: {process.stderr.decode()}"
                )

        if inferred_filename != output:
            inferred_filename.rename(output)



import enum
from pathlib import Path

from helix.configuration import MANAGER_CFG
from helix.reference.genome_metadata_loader import Genome
from helix.utility.external import External
from helix.files.file_type_checker import FileType, FileTypeChecker


class BgzipAction(enum.Enum):
    Compress = 0
    Decompress = 1
    Reindex = 2


class BGzip:
    def __init__(
        self,
        external: External = External(),
        file_type_checker=FileTypeChecker(),
        config=MANAGER_CFG.EXTERNAL,
    ) -> None:
        self._type_checker = file_type_checker
        self._external = external
        self._config = config

    def perform(self, genome: Genome, file: Path) -> Path:
        if self._type_checker.get_type(file) == FileType.BGZIP:
            if genome.fasta != file:
                if genome.fasta.exists():
                    genome.fasta.unlink()
                file.rename(genome.fasta)
            return genome.fasta
        return self.bgzip_wrapper(file, genome.fasta)

    def _gzip_filename(self, input: Path, action: BgzipAction):
        if action == BgzipAction.Compress:
            return Path(str(input) + ".gz")
        elif action == BgzipAction.Decompress:
            if len(input.suffixes) == 0:
                raise RuntimeError(
                    f"Unable to determine decompressed filename, invalid filename {str(input)} (no extensions)."
                )
            ext = "".join(input.suffixes[:-1])
            name = input.name.removesuffix("".join(input.suffixes))
            return input.with_name(name + ext)
        elif action == BgzipAction.Reindex:
            return Path(str(input) + ".gzi")
        else:
            raise RuntimeError(f"Action {action.name} not supported.")

    def bgzip_wrapper(
        self,
        input: Path,
        output: Path,
        action: BgzipAction = BgzipAction.Compress,
    ) -> Path:
        if output.exists():
            output.unlink()

        action_flags = {
            BgzipAction.Compress: "-if",
            BgzipAction.Decompress: "-d",
            BgzipAction.Reindex: "-r",
        }
        inferred_filename = self._gzip_filename(input, action)

        out = self._external.bgzip(
            [action_flags[action], str(input), "-@", str(self._config.threads)],
            wait=True,
        )
        if inferred_filename != output:
            inferred_filename.rename(output)
            if action == BgzipAction.Compress:
                target = Path(str(output) + ".gzi")
                inferred_gzi_filename = Path(str(inferred_filename) + ".gzi")
                if target.exists():
                    target.unlink()
                if not inferred_gzi_filename.exists():
                    raise FileNotFoundError(
                        f"BGZIP index not found for {inferred_gzi_filename.name}"
                    )
                inferred_gzi_filename.rename(target)
        return output



import enum
import logging
import shlex
import subprocess

from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.configuration import MANAGER_CFG
from helix.utility.external import External


class VariantCallingType(enum.Enum):
    InDel = enum.auto()
    SNP = enum.auto()
    Both = enum.auto()


class VariantCaller:
    def __init__(
        self,
        repo_config=MANAGER_CFG.REPOSITORY,
        ext_config=MANAGER_CFG.EXTERNAL,
        external: External = External(),
        progress=None,
        logger=logging.getLogger(__name__),
    ) -> None:
        self._external = external
        self._ext_config = ext_config
        self._ploidy = str(repo_config.metadata.joinpath("ploidy.txt"))
        self._is_quitting = False
        self._current_operation = None
        self._progress = progress
        self._logger = logger

    def call(self, aligned_file: AlignmentMapFile, type=VariantCallingType.Both):
        if aligned_file.file_info.index_stats is None:
            raise RuntimeError("Index stats cannot be None for variant calling")
        if aligned_file.file_info.reference_genome.ready_reference is None:
            raise RuntimeError("Reference cannot be None for variant calling")
        self.current_file = aligned_file
        reference = str(aligned_file.file_info.reference_genome.ready_reference.fasta)
        self._logger.info(
            f"Calling variants with {aligned_file.file_info.reference_genome.ready_reference}"
        )
        input_file = aligned_file.path

        skip_variant_opt = ""
        if type == VariantCallingType.InDel:
            skip_variant_opt = "-V snps"
        elif type == VariantCallingType.SNP:
            skip_variant_opt = "-V indels"

        output_file = aligned_file.path.with_name(
            f"{aligned_file.path.stem}_{type.name}.vcf.gz"
        )

        pileup_opt = f'mpileup -B -I -C 50 --threads {self._ext_config.threads} -f "{reference}" -Ou "{input_file}"'
        call_opt = f'call --ploidy-file "{self._ploidy}" {skip_variant_opt} -mv -P 0 --threads {self._ext_config.threads} -Oz -o "{output_file}"'
        tabix_opt = f'-p vcf "{output_file}"'

        pileup_opt = shlex.split(pileup_opt)
        call_opt = shlex.split(call_opt)
        tabix_opt = shlex.split(tabix_opt)

        progress = None
        if self._progress is not None:
            mapped_sequences = [
                x.mapped for x in self.current_file.file_info.index_stats
            ]
            mapped_sequences = sum(mapped_sequences)

            # 112 bytes seems to be the average bytes written for a single base
            total_bytes = (
                mapped_sequences
                * self.current_file.file_info.alignment_stats.average_length
                * 112
            )

            progress = BaseProgressCalculator(
                self._progress, total_bytes, "[1/2] Calling"
            )
            progress = progress.compute_on_write_bytes

        self._current_operation = self._external.bcftools(
            pileup_opt,
            stdout=subprocess.PIPE,
            io=progress,
        )

        call = self._external.bcftools(call_opt, stdin=self._current_operation.stdout)
        call.communicate()
        if self._is_quitting:
            return

        progress = BaseProgressCalculator(
            self._progress, output_file.stat().st_size, "[2/2] Indexing"
        )
        self._current_operation = self._external.tabix(
            tabix_opt, wait=True, io=progress.compute_on_read_bytes
        )

    def kill(self):
        self._is_quitting = True
        operation = self._current_operation
        try:
            if operation is not None:
                operation.kill()
        except Exception as e:
            self._logger.error(f"Error while killing variant calling: {e!s}")



import datetime

from jinja2 import Environment, FileSystemLoader

from helix.adapters.alignment_map_file_info_adapter import AlignmentMapFileInfoAdapter
from helix.adapters.alignment_stats_adapter import AlignmentStatsAdapter
from helix.adapters.header_adapter import HeaderAdapter
from helix.adapters.index_stats_adapter import IndexStatsAdapter
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.configuration import MANAGER_CFG
from helix.data.tabular_data import TabularData
from helix.VERSION import version


class HTMLAlignedFileReport:
    def __init__(self, config=MANAGER_CFG.REPOSITORY) -> None:
        self._templates_folder = config.metadata.joinpath("report_templates")

    def from_aligned_file(file=AlignmentMapFile):
        file_info_adapted = AlignmentMapFileInfoAdapter.adapt(file.file_info)
        index_stat_adapted = IndexStatsAdapter.adapt(file.file_info.index_stats)
        alignment_stats_adapted = AlignmentStatsAdapter.adapt(
            file.file_info.alignment_stats
        )
        header_adapted = HeaderAdapter.adapt(file.header)

        html_report = HTMLAlignedFileReport()
        return html_report.make(
            file_info_adapted,
            index_stat_adapted,
            alignment_stats_adapted,
            header_adapted.sequences,
            header_adapted.read_groups,
            header_adapted.programs,
            header_adapted.comments,
        )

    def make(
        self,
        file_info_data: TabularData,
        index_stat_data: TabularData,
        alignment_stats_data: TabularData,
        sequences_data: TabularData,
        read_groups_data: TabularData,
        programs_data: TabularData,
        comments_data: TabularData,
    ):
        file_loader = FileSystemLoader(self._templates_folder)
        env = Environment(loader=file_loader)
        template = env.get_template("table_tab.html")
        return template.render(
            file_info_data=file_info_data,
            index_stat_data=index_stat_data,
            alignment_stats_data=alignment_stats_data,
            sequences_data=sequences_data,
            read_groups_data=read_groups_data,
            programs_data=programs_data,
            comments_data=comments_data,
            general={"now": datetime.datetime.now(), "version": version},
        )



import subprocess
import time

from helix.data.coverage_stats import CoverageStats, DepthBin
from helix.alignment_map.alignment_map_file import AlignmentMapFile
from helix.data.file_type import FileType
from helix.data.sequence_type import SequenceType
from helix.fasta.reference import ReferenceStatus
from helix.naming.converter import Converter
from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.utility.external import External
from helix.utility.regions import RegionType, Regions


class CoverageStatsCalculator:
    def __init__(self, external=External(), regions=Regions(), progress=None) -> None:
        self._external = external
        self._progress = progress
        self._regions = regions
        self._progress_calc = None

    def get_stats(self, file: AlignmentMapFile, region: RegionType = None):
        options = ["depth"]
        if region is not None:
            bed_path = self._regions.get_path(
                file.file_info.reference_genome.ready_reference.build,
                region,
                file.header.sequences.values(),
            )
            options.extend(["-b", str(bed_path)])
        else:
            options.append("-aa")

        if file.file_info.file_type == FileType.CRAM:
            if file.file_info.reference_genome.status != ReferenceStatus.Available:
                raise FileNotFoundError(
                    "Unable to find the reference genome for loaded file."
                )
            options.extend(["-T", file.file_info.reference_genome.ready_reference])

        options.append(str(file.path))

        if self._progress is not None:
            total_bases = sum(
                [
                    x.length
                    for x in file.header.sequences.values()
                    if x.type != SequenceType.Unmapped
                ]
            )
            self._progress_calc = BaseProgressCalculator(
                self._progress, total_bases, "Calculating depth"
            )
        process = self._external.samtools(options, stdout=subprocess.PIPE, text=True)
        return self.analyze_depth_lines(iter(process.stdout.readline, ""), file)

    def analyze_depth_lines(self, lines, file: AlignmentMapFile):
        # Stats are organized in bins according to how many times
        # a specific position was read.
        entries_by_name = {}
        sums_by_name = {}

        for name, sequence in file.header.sequences.items():
            if sequence.type == SequenceType.Other:
                continue
            entries_by_name[name] = [0] * 4
            sums_by_name[name] = [0] * 4

        # Format is: sequence name, position (unused), # reads
        last_time = time.time()
        line_counter = 0
        for line in lines:
            if self._progress_calc is not None:
                now = time.time()
                if (now - last_time) > 1:
                    self._progress_calc.compute(line_counter)
                    last_time = now
                line_counter += 1
            line = line.split()
            name = line[0]
            reads = int(line[2])

            if name not in entries_by_name:
                continue

            if reads > 7:
                entries_by_name[name][DepthBin.MoreThan7] += 1
                sums_by_name[name][DepthBin.MoreThan7] += reads
            elif reads > 3:
                entries_by_name[name][DepthBin.Between3And7] += 1
                sums_by_name[name][DepthBin.Between3And7] += reads
            elif reads > 0:
                entries_by_name[name][DepthBin.Between0And3] += 1
                sums_by_name[name][DepthBin.Between0And3] += reads
            elif reads == 0:
                entries_by_name[name][DepthBin.Zero] += 1

        statistics = []
        for name in entries_by_name:
            current = CoverageStats()
            current.sequence_name = Converter.canonicalize(name)
            current.bin_entries = entries_by_name[name]
            current.bin_sum = sums_by_name[name]

            zero_count = current.bin_entries[DepthBin.Zero]
            non_zero_count = sum(
                [
                    current.bin_entries[DepthBin.Between0And3],
                    current.bin_entries[DepthBin.Between3And7],
                    current.bin_entries[DepthBin.MoreThan7],
                ]
            )

            all_sums = sum(current.bin_sum)
            all_count = zero_count + non_zero_count

            current.non_zero_percentage = non_zero_count / (all_count + 1)
            current.non_zero_average = 0
            if non_zero_count != 0:
                current.non_zero_average = all_sums / non_zero_count

            current.all_average = all_sums / (all_count + 1)
            statistics.append(current)
        return statistics



from helix.configuration import MANAGER_CFG
from helix.data.mitochondrial_model_type import MitochondrialModelType
from helix.data.sequence import Sequence


class MitochondrialModel:
    def __init__(self, sequence, url, type, config=MANAGER_CFG.REPOSITORY) -> None:
        self.sequence: Sequence = sequence
        self.url = url
        self.type: MitochondrialModelType = type
        self.path = config.mtdna.joinpath(self.type.name + ".fasta")

    def get_data(self):
        with self.path.open("r") as f:
            return f.read()

    def __str__(self) -> str:
        return f"{self.type.name}"

    def __repr__(self) -> str:
        return f"{self.type.name}"



from helix.data.chromosome_name_type import ChromosomeNameType
from helix.data.sequence_type import SequenceType
from helix.naming.lookup_tables import (
    GENBANK_T2T_TO_NUMBER,
    GENBANK_TO_NUMBER,
    NUMBER_TO_GENBANK,
    NUMBER_TO_GENBANK_T2T,
    NUMBER_TO_REFSEQ,
    NUMBER_TO_REFSEQ_T2T,
    REFSEQ_T2T_TO_NUMBER,
    REFSEQ_TO_NUMBER,
)


class Converter:
    def canonicalize(sequence_name: str) -> str:
        """Convert a sequence name into a "canonical" form,
        which is essentially the Number format:
        - Digits only for autosome
        - X/Y for sexual
        - M for mitochondrial
        - Other sequences are not touched

        Args:
            sequence_name (str): Sequence to convert

        Returns:
            str: Converted name sequence.
        """
        return Converter.convert(sequence_name, ChromosomeNameType.Number)

    def get_type(sequence_name: str) -> SequenceType:
        canonical_name = Converter.canonicalize(sequence_name)
        if canonical_name.isnumeric():
            return SequenceType.Autosome
        elif canonical_name == "X":
            return SequenceType.X
        elif canonical_name == "Y":
            return SequenceType.Y
        elif canonical_name == "M":
            return SequenceType.Mitochondrial
        elif canonical_name == "*":
            return SequenceType.Unmapped
        return SequenceType.Other

    def _find_in_table(input, table):
        normalized = input
        for key in table:
            if input.startswith(key):
                normalized = table[key]
                break
        return normalized

    def convert(input: str, target: ChromosomeNameType) -> str:
        # Chr to Number (e.g., chr1 -> 1; chrMT->MT)
        normalized = input.upper()
        if normalized.startswith("CHR"):
            normalized = normalized.replace("CHR", "", 1)

        # MT->M. This accounts also for the translation chrMT -> M,
        # since it's executed after the previous if.
        if normalized.startswith("MT"):
            normalized = normalized.replace("MT", "M", 1)

        #  Accession to Number (e.g., NC_000001 -> 1)
        if input.startswith("NC_"):
            normalized = Converter._find_in_table(normalized, REFSEQ_TO_NUMBER)
            normalized = Converter._find_in_table(normalized, REFSEQ_T2T_TO_NUMBER)
        elif input.startswith("CM") or input.startswith("J"):
            normalized = Converter._find_in_table(normalized, GENBANK_TO_NUMBER)
        elif input.startswith("CP"):
            normalized = Converter._find_in_table(normalized, GENBANK_T2T_TO_NUMBER)

        # Convert from Number format to target
        normalized = normalized.upper()
        if target == ChromosomeNameType.Chr:
            if normalized.isnumeric() or normalized in ["M", "X", "Y"]:
                return "chr" + normalized.upper()
            return normalized
        elif target == ChromosomeNameType.Number:
            return normalized
        elif target == ChromosomeNameType.GenBank:
            return NUMBER_TO_GENBANK.get(normalized, None)
        elif target == ChromosomeNameType.RefSeq:
            return NUMBER_TO_REFSEQ.get(normalized, None)
        elif target == ChromosomeNameType.RefSeqT2T:
            return NUMBER_TO_REFSEQ_T2T.get(normalized, None)
        elif target == ChromosomeNameType.GenBankT2T:
            return NUMBER_TO_GENBANK_T2T.get(normalized, None)
        raise ValueError(f"Converting to unrecognized target format: {target.name}")



import time


class BaseProgressCalculator:
    """Provide an abstraction over ProcessIOMonitor to compute an ETA
    and a percentage for a long operation (i.e., variant calling, indexing, etc.)
    """

    def __init__(
        self,
        progress,
        total_bytes,
        op_name="",
    ) -> None:
        self._previous_bytes = None
        self._previous_time = None
        self._bytes_per_second = [None] * 100
        self._average_writes_index = 0
        self._op_name = op_name
        self._progress = progress
        self._total_bytes = total_bytes
        self.compute_on_write_bytes = lambda _, w: self.compute(w)
        self.compute_on_read_bytes = lambda r, _: self.compute(r)

    def compute(self, current_bytes):
        if current_bytes is None:
            self._progress(None, None)
            return
        current_time = time.time()
        delta_time = 0
        delta_bytes = 0
        if self._previous_bytes is not None and self._previous_time is not None:
            delta_bytes = current_bytes - self._previous_bytes
            delta_time = current_time - self._previous_time
        else:
            self._previous_bytes = current_bytes
            self._previous_time = current_time
            return
        # Let's say the delta is null, just don't update anything and
        # let's try on next iteration to get some proper stats.
        # Delta time and delta bytes should never decrease, so we'll
        # have another change to compute stats later.
        if delta_bytes == 0 or delta_time == 0:
            return

        self._previous_bytes = current_bytes
        self._previous_time = current_time

        current_bytes_to_completion = self._total_bytes - current_bytes
        current_bytes_per_second = delta_bytes / delta_time
        if self._average_writes_index == len(self._bytes_per_second):
            self._average_writes_index = 0

        self._bytes_per_second[self._average_writes_index] = current_bytes_per_second
        self._average_writes_index += 1
        non_none_bytes_per_second = [x for x in self._bytes_per_second if x is not None]
        average_bytes_per_second = None
        if len(non_none_bytes_per_second) == 0:
            return
        average_bytes_per_second = sum(non_none_bytes_per_second) / len(
            non_none_bytes_per_second
        )
        if average_bytes_per_second == 0:
            return

        second_to_completion = current_bytes_to_completion / average_bytes_per_second

        # Set a threshold to not have the ETA jumping around
        threshold = len(non_none_bytes_per_second) < 10
        if threshold:
            formatted_delta = "Calculating, please stand-by."
        else:
            if second_to_completion != 0:
                hours, remainder = divmod(second_to_completion, 3600)
                minutes, seconds = divmod(remainder, 60)
                formatted_delta = f"{int(hours):02}:{int(minutes):02}:{int(seconds):02}"
            else:
                formatted_delta = "Unavailable"

        percentage = current_bytes * 100 / self._total_bytes

        formatted_label = f"{self._op_name}, ETA: {formatted_delta}"
        self._progress(formatted_label, percentage)



from helix.data.chromosome_name_type import ChromosomeNameType
from helix.naming.converter import Converter
from helix.naming.lookup_tables import ACCESSION_TO_NUMBER


def test_convert_to_number():
    converted_chr1 = Converter.convert("chr1", ChromosomeNameType.Number)
    converted_1 = Converter.convert("1", ChromosomeNameType.Number)
    converted_chrmt = Converter.convert("chrMT", ChromosomeNameType.Number)
    converted_mt = Converter.convert("MT", ChromosomeNameType.Number)
    converted_m = Converter.convert("M", ChromosomeNameType.Number)
    converted_chrx = Converter.convert("chrX", ChromosomeNameType.Number)
    converted_x = Converter.convert("X", ChromosomeNameType.Number)
    converted_chry = Converter.convert("chrY", ChromosomeNameType.Number)
    converted_y = Converter.convert("Y", ChromosomeNameType.Number)
    converted_acc = Converter.convert("CM000663", ChromosomeNameType.Number)

    assert converted_chr1 == "1"
    assert converted_1 == "1"
    assert converted_acc == "1"
    assert converted_chrmt == "M"
    assert converted_mt == "M"
    assert converted_m == "M"
    assert converted_chrx == "X"
    assert converted_x == "X"
    assert converted_chry == "Y"
    assert converted_y == "Y"


def test_convert_to_chr():
    converted_chr1 = Converter.convert("chr1", ChromosomeNameType.Chr)
    converted_1 = Converter.convert("1", ChromosomeNameType.Chr)
    converted_chrmt = Converter.convert("chrMT", ChromosomeNameType.Chr)
    converted_mt = Converter.convert("MT", ChromosomeNameType.Chr)
    converted_m = Converter.convert("M", ChromosomeNameType.Chr)
    converted_chrx = Converter.convert("chrX", ChromosomeNameType.Chr)
    converted_x = Converter.convert("X", ChromosomeNameType.Chr)
    converted_chry = Converter.convert("chrY", ChromosomeNameType.Chr)
    converted_y = Converter.convert("Y", ChromosomeNameType.Chr)
    converted_acc = Converter.convert("CM000663", ChromosomeNameType.Chr)

    assert converted_chr1 == "chr1"
    assert converted_1 == "chr1"
    assert converted_acc == "chr1"
    assert converted_chrmt == "chrM"
    assert converted_mt == "chrM"
    assert converted_m == "chrM"
    assert converted_chrx == "chrX"
    assert converted_x == "chrX"
    assert converted_chry == "chrY"
    assert converted_y == "chrY"


def test_convert_to_accession():
    converted_chr1 = Converter.convert("chr1", ChromosomeNameType.GenBank)
    converted_1 = Converter.convert("1", ChromosomeNameType.GenBank)
    converted_chrmt = Converter.convert("chrMT", ChromosomeNameType.GenBank)
    converted_mt = Converter.convert("MT", ChromosomeNameType.GenBank)
    converted_m = Converter.convert("M", ChromosomeNameType.GenBank)
    converted_chrx = Converter.convert("chrX", ChromosomeNameType.GenBank)
    converted_x = Converter.convert("X", ChromosomeNameType.GenBank)
    converted_chry = Converter.convert("chrY", ChromosomeNameType.GenBank)
    converted_y = Converter.convert("Y", ChromosomeNameType.GenBank)

    assert converted_chr1 == "CM000663"
    assert converted_1 == "CM000663"
    assert converted_chrmt == "J01415"
    assert converted_mt == "J01415"
    assert converted_m == "J01415"
    assert converted_chrx == "CM000685"
    assert converted_x == "CM000685"
    assert converted_chry == "CM000686"
    assert converted_y == "CM000686"


def test_lookup_values_uniqueness():
    for table in ACCESSION_TO_NUMBER:
        assert len(set(x for x in table.values())) == len(table)



import pytest

from helix.fasta.letter_run_collection import LetterRunCollection


def test_sequence_with_invalid_length():
    with pytest.raises(IndexError) as e:
        LetterRunCollection("Foo", 0)
    assert "greater than" in str(e.value)


def test_end_incomplete_sequence():
    with pytest.raises(ValueError) as e:
        sut = LetterRunCollection("Foo", 100)
        sut.end(99)
    assert "but processed" in str(e.value)


def test_open_run_on_ended_sequence():
    with pytest.raises(RuntimeError) as e:
        sut = LetterRunCollection("Foo", 100)
        sut.end(100)
        sut.open_run(101)
    assert "ended" in str(e.value)


def test_double_open_run():
    with pytest.raises(RuntimeError) as e:
        sut = LetterRunCollection("Foo", 10)
        sut.open_run(0)
        sut.open_run(0)
    assert "already opened" in str(e.value)


def test_negative_open():
    with pytest.raises(IndexError) as e:
        sut = LetterRunCollection("Foo", 10)
        sut.open_run(-10)
    assert "negative" in str(e.value)


def test_close_run_less_equal_than_open():
    with pytest.raises(RuntimeError) as e:
        sut = LetterRunCollection("Foo", 10)
        sut.open_run(0)
        sut.close_run(0)
    assert "greater" in str(e.value)


def test_only_close_run():
    with pytest.raises(RuntimeError) as e:
        sut = LetterRunCollection("Foo", 10)
        sut.close_run(0)
    assert "already closed" in str(e.value)


def test_double_close_run():
    with pytest.raises(RuntimeError) as e:
        sut = LetterRunCollection("Foo", 1)
        sut.open_run(0)
        sut.close_run(1)
        sut.close_run(1)
    assert "already closed" in str(e.value)


def test_close_run_greater_than_open():
    # Arrange
    sut = LetterRunCollection("Foo", 10)
    # Act
    sut.open_run(0)
    sut.close_run(1)
    # Assert
    assert len(sut.runs) == 1
    assert sut.runs[0].start == 0
    assert sut.runs[0].length == 1


def test_overlapping_runs():
    with pytest.raises(ValueError) as e:
        sut = LetterRunCollection("Foo", 19)
        sut.open_run(0)
        sut.close_run(15)

        sut.open_run(15)
        sut.close_run(19)
    assert "overlapping" in str(e.value)


def test_filter():
    # Arrange
    sequence = LetterRunCollection("Foo", 449)
    # Act
    sequence.open_run(0)
    sequence.close_run(150)
    sequence.open_run(300)
    sequence.close_run(449)
    sut = sequence.filter(lambda x: x.length > 149)
    # Assert
    assert len(sut) == 1
    assert sut[0].start == 0
    assert sut[0].length == 150


def test_is_run_open():
    sut = LetterRunCollection("Foo", 150)
    sut.open_run(0)
    assert sut.is_run_open() == True
    sut.close_run(150)
    assert sut.is_run_open() == False



from jinja2 import Environment, FileSystemLoader

from helix.configuration import MANAGER_CFG
from helix.data.tabular_data import TabularData


class HTMLSimpleTableReport:
    def __init__(self, name, config=MANAGER_CFG.REPOSITORY) -> None:
        self._name = name
        self._templates_folder = config.metadata.joinpath("report_templates")

    def make(self, table: TabularData):
        file_loader = FileSystemLoader(self._templates_folder)
        env = Environment(loader=file_loader)
        template = env.get_template("table_simple.html")
        return template.render(data={"title": self._name, "table": table})



import configparser
import logging
import multiprocessing
from pathlib import Path

# Why third_party is not recognized by the IDE?
# See helix/__init__.py
from helix import metadata, mtDNA, third_party

logger = logging.getLogger("configuration")
logging.getLogger().setLevel(logging.DEBUG)


class HelixDefaults:
    """Specify some directory where to find configuration files."""

    HELIX_FOLDER = Path(__file__).parents[1]
    LOCAL_FOLDER = Path(Path.home(), ".helix")
    LOCAL_CONFIG = Path(HELIX_FOLDER, "configuration", "main.ini")
    GLOBAL_CONFIG = Path(Path.home(), ".helix", "main.ini")


class GeneralConfig:
    """General configuration for the app.

    Every attribute of this class can be loaded from the configuration
    files if present. See the documentation for ConfigurationManager.
    NOTE: you can add every type of variables as long as they can be
    constructed with a (single) string. Otherwise this will fail.

    Attributes:
        last_path (Path): path for the last file opened
        log_level (str): minimum log level. Possible values: DEBUG,
            INFO, WARNING, CRITICAL
    """

    def __init__(self) -> None:
        self.last_path: Path = Path.home()
        self.log_level: str = "DEBUG"


class ExternalConfig:
    """Configuration for 3rd party dependencies.

    Every attribute of this class can be loaded from the configuration
    files if present. See the documentation for ConfigurationManager.
    NOTE: you can add every type of variables as long as they can be
    constructed with a (single) string. Otherwise this will fail.

    Attributes:
        root (Path): folder that contains 3rd party dependencies.
            If this does not point to a valid directory, the files
            should be under PATH.
        threads (int): How many threads is possible to use.

    """

    def __init__(self) -> None:
        self.root: Path = Path(third_party.__file__).parent
        self.threads: int = multiprocessing.cpu_count()


class RepositoryConfig:
    """Configuration for reference genome and metadata repository.

    Every attribute of this class can be loaded from the configuration
    files if present. See the documentation for ConfigurationManager.
    NOTE: you can add every type of variables as long as they can be
    constructed with a (single) string. Otherwise this will fail.

    Attributes:
        repository (Path): Root folder for reference genomes.
        temporary (Path): Temporary files directory.
        metadata (Path): Root folder for metadata.
        mtdna (Path): Root folder for mtDNA files.
    """

    def __init__(self) -> None:
        self.genomes: Path = Path(HelixDefaults.LOCAL_FOLDER, "genomes")
        self.temporary: Path = Path(HelixDefaults.LOCAL_FOLDER, "temp")
        self.log_path: Path = Path(HelixDefaults.LOCAL_FOLDER, "logs")
        self.metadata: Path = Path(metadata.__file__).parent
        self.mtdna: Path = Path(mtDNA.__file__).parent


class AlignmentStatsConfig:
    """Configuration for alignment stats calculation.

    Every attribute of this class can be loaded from the configuration
    files if present. See the documentation for ConfigurationManager.
    NOTE: you can add every type of variables as long as they can be
    constructed with a (single) string. Otherwise this will fail.

    Attributes:
        skip (int): How many samples to skip at the beginning
            of the file (where lots of repetitive and low quality
            sequences usually are).
        samples (int): How many samples to consider when calculating
            the stats.
    """

    def __init__(self) -> None:
        self.skip: int = 40000
        self.samples: int = 20000


class ConfigurationManager:
    """Initialize the configuration.

    Every static attribute declared here will be initialized
    when the class is initialized by looking for sections
    with the same as the attribute, containing the same attributes
    as the attribute.

    Attributes:
        GENERAL (GeneralConfig): Configuration parameters for the whole
            app.
        EXTERNAL (ExternalConfig): Configuration for 3rd party dependencies.
        REPOSITORY (RepositoryConfig): Configuration for reference genome
            and metadata repository.
        ALIGNMENT_STATS (AlignmentStatsConfig): Configuration for alignment
            stats calculation.
    """

    GENERAL: GeneralConfig = GeneralConfig()
    EXTERNAL: ExternalConfig = ExternalConfig()
    REPOSITORY: RepositoryConfig = RepositoryConfig()
    ALIGNMENT_STATS: AlignmentStatsConfig = AlignmentStatsConfig()

    def __init__(self) -> None:
        self.load()

    def load(self) -> None:
        self._parser = configparser.ConfigParser(interpolation=None)
        if HelixDefaults.LOCAL_CONFIG.exists():
            logging.info(f"Loading {HelixDefaults.LOCAL_CONFIG}")
        if HelixDefaults.GLOBAL_CONFIG.exists():
            logging.info(f"Loading {HelixDefaults.GLOBAL_CONFIG}")
        self._parser.read(HelixDefaults.LOCAL_CONFIG)
        self._parser.read(HelixDefaults.GLOBAL_CONFIG)

        for var_name, var_value in ConfigurationManager.__dict__.items():
            if var_name.startswith("__"):
                continue
            if getattr(var_value, "__dict__") is None:
                continue
            if type(var_value) is type(ConfigurationManager.load):
                continue

            section = var_name.lower()
            if section not in self._parser:
                continue
            for key, value in self._parser[section].items():
                if key not in var_value.__dict__:
                    logging.warning(f"Configuration {section}.{key} not known")
                    continue
                if value is None:
                    continue
                var_type = type(var_value.__dict__[key])
                var_value.__dict__[key] = var_type(value)

    def save(self, save_defaults=False):
        for var_name, var_value in ConfigurationManager.__dict__.items():
            if var_name.startswith("__"):
                continue
            if getattr(var_value, "__dict__") is None:
                continue
            if type(var_value) is type(ConfigurationManager.load):
                continue

            default_item = type(var_value)()
            section = var_name.lower()
            if section not in self._parser:
                self._parser.add_section(section)
            for key, value in var_value.__dict__.items():
                if default_item.__dict__[key] != value or save_defaults:
                    self._parser[section][key] = str(value)
            if len(self._parser[section]) == 0 and not save_defaults:
                self._parser.remove_section(section)
        ordered_config_paths = [HelixDefaults.GLOBAL_CONFIG, HelixDefaults.LOCAL_CONFIG]
        if not any([x.exists() for x in ordered_config_paths]):
            with ordered_config_paths[1].open("wt") as f:
                pass
        for config in ordered_config_paths:
            if not config.exists():
                continue
            with config.open("wt") as f:
                self._parser.write(f)
            break


MANAGER_CFG = None
if MANAGER_CFG is None:
    HelixDefaults.GLOBAL_CONFIG.parent.mkdir(parents=True, exist_ok=True)
    HelixDefaults.GLOBAL_CONFIG.touch(exist_ok=True)
    MANAGER_CFG = ConfigurationManager()
    MANAGER_CFG.REPOSITORY.temporary.mkdir(parents=True, exist_ok=True)
    MANAGER_CFG.REPOSITORY.genomes.mkdir(parents=True, exist_ok=True)
    MANAGER_CFG.REPOSITORY.mtdna.mkdir(parents=True, exist_ok=True)



import sentry_sdk

import helix.gui
from helix.utility.check_prerequisites import CheckPrerequisites


def main():
    sentry_sdk.init(
        dsn="https://7942f481884f55dfc85350ef336b3299@o4507282907856896.ingest.de.sentry.io/4507282933874768",
        traces_sample_rate=1.0,
        profiles_sample_rate=1.0,
    )
    CheckPrerequisites().check_prerequisites()
    helix.gui.main()


if __name__ == "__main__":
    main()



import logging
from collections import OrderedDict

from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.configuration import RepositoryConfig
from helix.data.genome import Genome
from helix.data.sequence import Sequence
from helix.fasta.reference import Reference
from helix.files.bgzip import BGzip, BgzipAction
from helix.files.decompressor import Decompressor
from helix.files.downloader import Downloader
from helix.reference.genome_metadata_loader import MetadataLoader
from helix.mtDNA.mt_dna import MtDNA
from helix.utility.samtools import Samtools


class Repository:
    def __init__(
        self,
        loader: MetadataLoader = MetadataLoader(),
        downloader: Downloader = Downloader(),
        mtdna: MtDNA = MtDNA(),
        config: RepositoryConfig = RepositoryConfig(),
    ) -> None:
        self._loader = loader
        self._downloader = downloader
        self._mtdna = mtdna
        self.genomes = self._loader.load()
        # Index individual sequences by their MD5 and length
        self._sequences_by_md5: dict[str, list[Sequence]] = self._group_sequences(
            lambda s: s.md5, self.genomes
        )
        self._sequences_by_length: dict[int, list[Sequence]] = self._group_sequences(
            lambda s: s.length, self.genomes
        )
        self._config = config

    def analyze_references(self):
        # Analyzing sequence by sequence
        different_md5_same_length = []
        for ln, seqs in self._sequences_by_length.items():
            unique_seq = set()
            name = None
            for seq in seqs:
                unique_seq.add(seq.md5)
                name = seq.name
            # Filtering by an arbitrary value so that what's left are mostly
            # normal chromosomes sequences.
            if (len(unique_seq) > 1) and ln > 57227414:
                different_md5_same_length.append(ln)
                logging.debug(
                    f"There are {len(unique_seq)} different MD5 for length {ln} ({name})."
                )

        # Analyzing the whole genome, looking for identical length sequences
        # with different MD5 sequences
        different_md5s_same_lengths = {}
        for genome in self.genomes:
            if genome.sequences is None:
                continue
            seq_lengths = ",".join([str(x.length) for x in genome.sequences])
            seq_md5 = ",".join([x.md5 for x in genome.sequences])
            if seq_lengths not in different_md5s_same_lengths:
                different_md5s_same_lengths[seq_lengths] = []
            if seq_md5 not in [x[0] for x in different_md5s_same_lengths[seq_lengths]]:
                different_md5s_same_lengths[seq_lengths].append((seq_md5, genome))

        ambiguous = len(
            [x for x, y in different_md5s_same_lengths.items() if len(y) > 1]
        )
        logging.debug(f"Found {ambiguous} possible ambiguous sequences of lengths:")
        index = 0
        for different_md5_same_length, md5s in different_md5s_same_lengths.items():
            if len(md5s) > 1:
                for md5, genome in md5s:
                    logging.debug(f"{index+1}) {genome.fasta_url!s}")
                index += 1

    def find(self, sequences: list[Sequence]):
        md5_available = all([x.md5 is not None for x in sequences])
        matching = OrderedDict([(x, []) for x in sequences])
        if md5_available:
            for sequence in sequences:
                if sequence.md5 in self._sequences_by_md5:
                    matching[sequence] = self._sequences_by_md5[sequence.md5]
        else:
            for sequence in sequences:
                if sequence.length in self._sequences_by_length:
                    matching[sequence] = self._sequences_by_length[sequence.length]
        return Reference(matching)

    def _group_sequences(self, criteria, genomes: list[Genome]):
        grouped = {}
        for genome in genomes:
            if genome.sequences is None:
                continue
            for sequence in genome.sequences:
                evaluated_criteria = criteria(sequence)
                if evaluated_criteria not in grouped:
                    grouped[evaluated_criteria] = []
                grouped[evaluated_criteria].append(sequence)
        return grouped

    def _get_sizes(self, genome: Genome):
        if genome.fai_url is None:
            fai_url = genome.fasta_url + ".fai"
            fai_size = self._downloader.get_file_size(fai_url)
            if fai_size is not None:
                genome.fai_url = fai_url

        if genome.gzi_url is None:
            gzi_url = genome.fasta_url + ".gzi"
            gzi_size = self._downloader.get_file_size(gzi_url)
            if gzi_size is not None:
                genome.gzi_url = gzi_url

        if genome.download_size is None:
            size = self._downloader.get_file_size(genome.fasta_url)
            if size is not None:
                genome.download_size = size
            else:
                raise RuntimeError(
                    f"Unable to get the size of the fasta file for {genome.fasta_url}"
                )

    def _get_sequences(self, genome: Genome):
        dictionary: AlignmentMapHeader = AlignmentMapHeader.load_from_file(genome.dict)
        sequences = []
        for name, entry in dictionary.sequences.items():
            sequence = Sequence(name, entry.length, entry.md5)
            sequences.append(sequence)
        return sequences

    def _create_companion_files(self, genome: Genome, force=False):
        samtools = Samtools()
        bgzip = BGzip()
        logging.info(f"{genome}: Starting post-download tasks.")
        if not genome.gzi.exists():
            logging.info(f"{genome}: Generating bgzip index.")
            bgzip.bgzip_wrapper(genome.fasta, genome.gzi, BgzipAction.Reindex)
        else:
            logging.info(f"{genome}: bgzip index exists.")

        if not genome.dict.exists():
            logging.info(f"{genome}: Generating dictionary file.")
            samtools.make_dictionary(genome.fasta, genome.dict)
        else:
            logging.info(f"{genome}: Dictionary file exists.")

    def self_test(self):
        for index, reference in enumerate(self.genomes):
            if reference.sequences is None:
                try:
                    logging.info(f"Processing {reference} as it has no sequences.")
                    genome = self.ingest(
                        reference.fasta_url, reference.source, reference.build
                    )
                    self.genomes[index] = genome
                    self._loader.save(self.genomes)
                except Exception as e:
                    logging.critical(e)

    def download(self, genome: Genome, progress=None, force=False):
        downloader = Downloader()
        decompressor = Decompressor()
        compressor = BGzip()

        if genome.fasta.exists() and not force:
            logging.info(f"File {genome.fasta.name} already exist. Re-using it.")
            self._create_companion_files(genome, force)
            return genome
        elif genome.fasta.exists() and force:
            genome.fasta.unlink()
            genome.bgzip_md5 = None
            genome.downloaded_md5 = None
            genome.decompressed_md5 = None
            genome.bgzip_size = None
            genome.download_size = None
            genome.decompressed_size = None

        logging.info(f"1/4: Downloading fasta from: {genome.fasta_url}.")
        download_output = downloader.perform(genome, progress)
        logging.info(f"2/4: Decompressing: {download_output.name}.")
        decompressor_output = decompressor.perform(genome, download_output)
        logging.info(f"3/4: Compressing to gzip: {decompressor_output.name}.")
        compressor_output = compressor.perform(genome, decompressor_output)
        logging.info(f"4/4: Extracting sequence data: {compressor_output.name}.")
        self._create_companion_files(genome)
        return genome

    def ingest(
        self, url: str = None, source: str = None, build: str = None, force=False
    ):
        genome = Genome(
            url, source=source, build=build, parent_folder=self._config.genomes
        )
        logging.info(f"Ingesting {genome}.")
        genome = self.download(genome, force)
        genome.sequences = self._get_sequences(genome)
        return genome


if __name__ == "__main__":
    import tempfile
    import hashlib
    from os import close
    import pathlib

    genomes = MetadataLoader().load()
    decompressor = BGzip()
    for genome in genomes:
        if genome.decompressed_md5 is None:
            continue
        if genome.fasta.exists():
            temp = tempfile.mkstemp(suffix=None, prefix=None, dir=None, text=False)
            close(temp[0])
            temp_file = pathlib.Path(temp[1])
            temp_file.unlink()
            decompressor.bgzip_wrapper(genome.fasta, temp_file, BgzipAction.Decompress)
            md5_hash = hashlib.md5()
            with temp_file.open("rb") as f:
                while True:
                    chunk = f.read(4096 * 1000)
                    if not chunk:
                        break
                    md5_hash.update(chunk)
            genome.decompressed_md5 = md5_hash.hexdigest()
            temp_file.unlink()
    MetadataLoader().save(genomes)
    exit()

    genomes = MetadataLoader().load()
    genomes = [x for x in genomes if x.decompressed_md5 is not None]
    for genome in genomes:
        target = genome.fasta.with_name(genome.decompressed_md5)
        target = target.with_suffix("".join(genome.fasta.suffixes))
        target_dict = genome.dict.with_name(genome.decompressed_md5)
        target_dict = target_dict.with_suffix("".join(genome.dict.suffixes))
        target_gzi = genome.gzi.with_name(genome.decompressed_md5)
        target_gzi = target_gzi.with_suffix("".join(genome.gzi.suffixes))
        if not target.exists() and genome.fasta.exists():
            genome.fasta.rename(target)
        if not target_dict.exists() and genome.dict.exists():
            genome.dict.rename(target_dict)
        if not target_gzi.exists() and genome.gzi.exists():
            genome.gzi.rename(target_gzi)



import os
import platform
from pathlib import Path

try:
    from win32com.client import Dispatch
except Exception:
    pass
import importlib.metadata


class Shortcut:
    def __init__(self) -> None:
        distribution = importlib.metadata.distribution("Helix")
        entry_point = distribution.entry_points["helix"]
        self.desktop = Path.home().joinpath("Desktop")
        self.path = None
        if len(entry_point.dist.files) > 0:
            self.path = entry_point.dist.locate_file(entry_point.dist.files[0])

    def _create_win(self):
        if self.path is None:
            return
        path = str(self.desktop.joinpath("Helix.lnk"))
        shell = Dispatch("WScript.Shell")
        shortcut = shell.CreateShortCut(path)
        shortcut.TargetPath = str(self.path)
        shortcut.WorkingDirectory = str(Path.home().joinpath("Documents"))
        shortcut.IconLocation = str(self.path)
        shortcut.Description = (
            "Genome file manipulation tool, https://github.com/Helix/Helix"
        )
        shortcut.save()
        return path

    def _create_mac(self):
        script = (
            'tell application "Finder"\n'
            + f'make alias file to POSIX file "{self.path!s}" at POSIX file "{self.desktop!s}"\n'
            + "end tell"
        )
        os.system(f"osascript -e '{script}'")
        return str(self.desktop)

    def _create_linux(self):
        if self.path is None:
            return
        shortcut_path = self.desktop.joinpath("Helix.desktop")
        desktop_entry = (
            "[Desktop Entry]\n"
            + "Type=Application\n"
            + "Name=Helix\n"
            + f"Exec={self.path!s}\n"
            + "Icon=\n"
            + "Terminal=false\n"
        )

        with shortcut_path.open("w") as file:
            file.write(desktop_entry)
        os.chmod(shortcut_path, 0o755)
        return str(shortcut_path)

    def create(self):
        os_type = platform.system()
        if os_type == "Windows":
            return self._create_win()
        elif os_type == "Linux":
            return self._create_linux()
        elif os_type == "Darwin":
            return self._create_mac()
        else:
            raise RuntimeError(f"Unsupported OS: {os_type}")



from enum import Enum, auto


class MicroarrayConverterExtensions(Enum):
    CSV = ".csv"
    TXT = ".txt"


class MicroarrayConverterTarget(Enum):
    All = auto()
    Ancestry_v1 = auto()
    Ancestry_v2 = auto()
    FTDNA_v1 = auto()
    FTDNA_v2 = auto()
    FTDNA_v3 = auto()
    LivingDNA_v1 = auto()
    LivingDNA_v2 = auto()
    MTHFR = auto()
    MyHeritage_v1 = auto()
    MyHeritage_v2 = auto()
    ReichLab_AADR = auto()
    ReichLab_HumanOrigins_v1 = auto()
    SelfDecode = auto()
    TellMeGen = auto()
    TwentyThreeAndMe_v3 = auto()
    TwentyThreeAndMe_v35 = auto()
    TwentyThreeAndMe_v4 = auto()
    TwentyThreeAndMe_v5 = auto()
    TwentyThreeAndMe_SNPs_API = auto()



from test.utility import MockFile, MockPath

import pytest

from helix.fasta.fasta_letter_counter import FASTALetterCounter


class MockGenome:
    def __init__(self, fasta_lines, dictionary_lines) -> None:
        self.fasta = MockPath(fasta_lines)
        self.dict = MockPath(dictionary_lines)


def gzip_open(path, mode):
    return MockFile(path.lines)


def test_run_continuing_across_lines_is_processed_correctly():
    # Arrange
    fa_lines = [
        ">1 irrelevant\n",
        f"{'N'*5}\n",
        f"{'N'*5}\n",
    ]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:10\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 0
    assert result[0].runs[0].length == 10


def test_run_starting_at_0_is_processed_correctly():
    # Arrange
    fa_lines = [">1 irrelevant\n", f"{'N'*3}{'A'*5}\n"]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:8\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 0
    assert result[0].runs[0].length == 3


def test_run_starting_in_the_middle_is_processed_correctly():
    # Arrange
    fa_lines = [
        ">1 irrelevant\n",
        f"{'A'*3}{'N'*5}{'A'*3}\n",
    ]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:11\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 3
    assert result[0].runs[0].length == 5


def test_run_ending_with_line_is_processed_correctly():
    # Arrange
    fa_lines = [">1 irrelevant", f"{'A'*3}{'N'*5}\n"]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:8\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 3
    assert result[0].runs[0].length == 5


def test_run_not_continuing_in_next_line_is_processed_correctly():
    # Arrange
    fa_lines = [
        ">1 irrelevant",
        f"{'A'*3}{'N'*5}\n",
        f"{'A'*3}\n",
    ]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:11\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 3
    assert result[0].runs[0].length == 5


def test_run_not_continuing_in_next_line_is_processed_correctly():
    # Arrange
    fa_lines = [
        ">1 irrelevant",
        f"{'A'*3}{'N'*5}\n",
        ">2 irrelevant",
        f"{'A'*3}{'N'*5}\n",
    ]
    dict_line = [
        "@HD\tVN:1.0\tSO:unsorted",
        "@SQ\tSN:1\tLN:8\tM5:dummy\tUR:file://c:/foo.fa.gz",
        "@SQ\tSN:2\tLN:8\tM5:dummy\tUR:file://c:/foo.fa.gz",
    ]
    pytest.MonkeyPatch().setattr("gzip.open", gzip_open)
    sut = FASTALetterCounter(MockGenome(fa_lines, dict_line))

    # Act
    result = sut.count_letters()

    # Assert
    assert len([x for x in result if x.name == "1"]) == 1
    assert len([x for x in result if x.name == "2"]) == 1
    assert len(result[0].runs) == 1
    assert result[0].runs[0].start == 3
    assert result[0].runs[0].length == 5
    assert len(result[1].runs) == 1
    assert result[1].runs[0].start == 3
    assert result[1].runs[0].length == 5



from helix.alignment_map.alignment_map_header import AlignmentMapHeader
from helix.data.sorting import Sorting


def test_header_parsed_correctly():
    lines = [
        "@HD\tVN:1.0\tSO:coordinate",
    ]
    sut = AlignmentMapHeader(lines)
    assert len(sut.sequences) == 0
    assert sut.metadata.sorted == Sorting.Coordinate
    assert sut.metadata.version == "1.0"


def test_unsorted_parsed_correctly():
    lines = [
        "@HD\tVN:1.0\tSO:unsorted",
    ]
    sut = AlignmentMapHeader(lines)
    assert len(sut.sequences) == 0
    assert sut.metadata.sorted == Sorting.Unsorted
    assert sut.metadata.version == "1.0"


def test_unsorted_parsed_correctly():
    lines = [
        "@HD\tVN:1.0\tSO:unknown",
    ]
    sut = AlignmentMapHeader(lines)
    assert len(sut.sequences) == 0
    assert sut.metadata.sorted == Sorting.Unknown
    assert sut.metadata.version == "1.0"


def test_sequences_parsed_correctly():
    lines = ["@SQ\tSN:Seq1\tLN:123\tM5:abcdefabcdef123123\tUR:/directory/genome.fasta"]
    sut = AlignmentMapHeader(lines)
    assert sut.metadata is None
    assert len(sut.sequences) == 1
    assert "Seq1" in sut.sequences
    assert sut.sequences["Seq1"].length == 123
    assert sut.sequences["Seq1"].md5 == "abcdefabcdef123123"
    assert sut.sequences["Seq1"].uri == "/directory/genome.fasta"



class MockFile:
    def __init__(self, lines) -> None:
        self.lines = lines

    def __enter__(self):
        return self.lines

    def __exit__(self, _, _1, _2):
        return


class MockPath:
    def __init__(
        self,
        lines=[],
        exists=True,
        name="foo.fa.gz",
        stem="foo.fa",
        parent="bar",
        suffix=".gz",
        files=[],
        **kwargs
    ) -> None:
        self.lines = lines
        self.name = name
        self.stem = stem
        self.parent = parent
        self.suffix = suffix
        self.files: list[MockPath] = files
        self._exists = exists
        self.__dict__.update(kwargs)

    def open(self, *_1):
        return MockFile(self.lines)

    def joinpath(self, path):
        name = self.name + "/" + path
        joined = [x for x in self.files if x.name == name]
        assert len(joined) == 1
        return joined[1]

    def exists(self):
        return self._exists



from pydantic import BaseModel


class AlignmentMapHeaderReadGroup(BaseModel):
    id: str = None
    barcode: str = None
    sequencing_center: str = None
    description: str = None
    date: str = None
    flow_order: str = None
    key_sequence: str = None
    library: str = None
    programs: str = None
    predicted_median_insert_size: int = None
    platform: str = None
    platform_model: str = None
    platform_unit: str = None
    sample: str = None



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'form.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    QTime,
    QUrl,
    Qt,
)
from PySide6.QtGui import (
    QAction,
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QGridLayout,
    QHBoxLayout,
    QHeaderView,
    QMainWindow,
    QMenu,
    QMenuBar,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QStatusBar,
    QTableWidget,
    QTableWidgetItem,
    QWidget,
)


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName("MainWindow")
        MainWindow.resize(724, 562)
        MainWindow.setMinimumSize(QSize(0, 20))
        self.actionSettings = QAction(MainWindow)
        self.actionSettings.setObjectName("actionSettings")
        self.actionOpen = QAction(MainWindow)
        self.actionOpen.setObjectName("actionOpen")
        self.actionExit = QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionDocumentation = QAction(MainWindow)
        self.actionDocumentation.setObjectName("actionDocumentation")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionLog_viewer = QAction(MainWindow)
        self.actionLog_viewer.setObjectName("actionLog_viewer")
        self.actionGenomes = QAction(MainWindow)
        self.actionGenomes.setObjectName("actionGenomes")
        self.actionCreate_launcher_on_desktop = QAction(MainWindow)
        self.actionCreate_launcher_on_desktop.setObjectName(
            "actionCreate_launcher_on_desktop"
        )
        self.actionHaplotype = QAction(MainWindow)
        self.actionHaplotype.setObjectName("actionHaplotype")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.extract = QPushButton(self.centralwidget)
        self.extract.setObjectName("extract")

        self.horizontalLayout.addWidget(self.extract)

        self.horizontalSpacer_2 = QSpacerItem(
            40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 3)

        self.progress = QProgressBar(self.centralwidget)
        self.progress.setObjectName("progress")
        sizePolicy = QSizePolicy(
            QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.progress.sizePolicy().hasHeightForWidth())
        self.progress.setSizePolicy(sizePolicy)
        self.progress.setMinimumSize(QSize(0, 0))
        self.progress.setValue(24)

        self.gridLayout.addWidget(self.progress, 2, 0, 1, 1)

        self.stop = QPushButton(self.centralwidget)
        self.stop.setObjectName("stop")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.stop.sizePolicy().hasHeightForWidth())
        self.stop.setSizePolicy(sizePolicy1)
        self.stop.setMinimumSize(QSize(0, 0))

        self.gridLayout.addWidget(self.stop, 2, 2, 1, 1)

        self.fileInformationTable = QTableWidget(self.centralwidget)
        self.fileInformationTable.setObjectName("fileInformationTable")
        sizePolicy2 = QSizePolicy(
            QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding
        )
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.fileInformationTable.sizePolicy().hasHeightForWidth()
        )
        self.fileInformationTable.setSizePolicy(sizePolicy2)
        self.fileInformationTable.setStyleSheet("")
        self.fileInformationTable.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.fileInformationTable.setAlternatingRowColors(False)
        self.fileInformationTable.setShowGrid(True)
        self.fileInformationTable.setGridStyle(Qt.DotLine)
        self.fileInformationTable.setSortingEnabled(False)
        self.fileInformationTable.setColumnCount(0)
        self.fileInformationTable.horizontalHeader().setVisible(False)
        self.fileInformationTable.horizontalHeader().setCascadingSectionResizes(False)
        self.fileInformationTable.horizontalHeader().setStretchLastSection(True)
        self.fileInformationTable.verticalHeader().setVisible(True)
        self.fileInformationTable.verticalHeader().setCascadingSectionResizes(False)
        self.fileInformationTable.verticalHeader().setMinimumSectionSize(30)
        self.fileInformationTable.verticalHeader().setDefaultSectionSize(30)
        self.fileInformationTable.verticalHeader().setHighlightSections(False)
        self.fileInformationTable.verticalHeader().setStretchLastSection(False)

        self.gridLayout.addWidget(self.fileInformationTable, 0, 0, 1, 3)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName("menubar")
        self.menubar.setGeometry(QRect(0, 0, 724, 21))
        sizePolicy3 = QSizePolicy(
            QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum
        )
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.menubar.sizePolicy().hasHeightForWidth())
        self.menubar.setSizePolicy(sizePolicy3)
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuTools = QMenu(self.menubar)
        self.menuTools.setObjectName("menuTools")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTools.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionSettings)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionExit)
        self.menuHelp.addAction(self.actionDocumentation)
        self.menuHelp.addAction(self.actionAbout)
        self.menuTools.addAction(self.actionLog_viewer)
        self.menuTools.addAction(self.actionGenomes)
        self.menuTools.addAction(self.actionHaplotype)
        self.menuTools.addSeparator()
        self.menuTools.addAction(self.actionCreate_launcher_on_desktop)

        self.retranslateUi(MainWindow)
        self.actionExit.triggered.connect(MainWindow.close)

        QMetaObject.connectSlotsByName(MainWindow)

    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(
            QCoreApplication.translate(
                "MainWindow", "Helix - Genome sequencing data manipulation tool", None
            )
        )
        self.actionSettings.setText(
            QCoreApplication.translate("MainWindow", "Settings", None)
        )
        self.actionOpen.setText(QCoreApplication.translate("MainWindow", "Open", None))
        self.actionExit.setText(QCoreApplication.translate("MainWindow", "Exit", None))
        self.actionDocumentation.setText(
            QCoreApplication.translate("MainWindow", "Documentation", None)
        )
        self.actionAbout.setText(
            QCoreApplication.translate("MainWindow", "About", None)
        )
        self.actionLog_viewer.setText(
            QCoreApplication.translate("MainWindow", "Log viewer", None)
        )
        self.actionGenomes.setText(
            QCoreApplication.translate("MainWindow", "Reference genomes", None)
        )
        self.actionCreate_launcher_on_desktop.setText(
            QCoreApplication.translate("MainWindow", "Create launcher on desktop", None)
        )
        self.actionHaplotype.setText(
            QCoreApplication.translate("MainWindow", "Haplotype", None)
        )
        self.extract.setText(QCoreApplication.translate("MainWindow", "Extract", None))
        self.stop.setText(QCoreApplication.translate("MainWindow", "Stop", None))
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", "File", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", "Help", None))
        self.menuTools.setTitle(QCoreApplication.translate("MainWindow", "Tools", None))

    # retranslateUi



import logging
import time
from subprocess import Popen
from threading import Thread

import debugpy
import psutil


class ProcessIOMonitor(Thread):
    def __init__(self, process: Popen, report, polling=1) -> None:
        super().__init__()
        self.process: Popen = process
        self.report = report
        self.polling = polling

    def run(self):
        if debugpy.is_client_connected():
            debugpy.debug_this_thread()
        try:
            process = psutil.Process(self.process.pid)
            while True:
                counters = process.io_counters()
                self.report(counters.read_bytes, counters.write_bytes)
                if process.status() != psutil.STATUS_RUNNING:
                    break
                time.sleep(self.polling)
        except Exception as e:
            # The process can end anytime and psutil will fail.
            # That's not an error.
            if not isinstance(e, psutil.NoSuchProcess):
                logging.error(f"Exception in ProcessIOMonitor: {e!s}")
            pass
        finally:
            # Signal the process ended.
            self.report(None, None)



import shutil
from pathlib import Path

import pefile


class ImportScanner:
    def __init__(self, paths: list[Path]):
        self.paths = paths

    def get_dlls(self, binary):
        to_scan = [binary]
        scanned = []

        dependencies = set([binary])
        while len(to_scan) > 0:
            file = to_scan.pop()
            pe = pefile.PE(str(file))
            for entry in pe.DIRECTORY_ENTRY_IMPORT:
                dependency = entry.dll.decode("utf-8")
                for path in self.paths:
                    target = Path(path, dependency)
                    if target.exists():
                        if target not in scanned and target not in to_scan:
                            to_scan.append(target)
                        dependencies.add(target)
            scanned.append(file)
        return dependencies


if __name__ == "__main__":
    """This script will scan certain executables and
    recursively look for dependencies.
    """

    msys2_additional_path = Path("/", "usr", "bin")

    files = [
        Path("bwa", "bwa.exe"),
        Path("htslib", "bgzip.exe"),
        Path("htslib", "tabix.exe"),
        Path("htslib", "htsfile.exe"),
        Path("samtools", "samtools.exe"),
        Path("bcftools", "bcftools.exe"),
        Path("minimap2", "minimap2.exe"),
    ]

    destination_root = Path("helix", "third_party")
    if not destination_root.exists():
        destination_root.mkdir(parents=True, exist_ok=True)

    destination_root.joinpath("__init__.py").touch()

    for file in files:
        scanner = ImportScanner([msys2_additional_path])
        binaries = scanner.get_dlls(file)

        for source_binary in binaries:
            destination_binary = destination_root.joinpath(source_binary.name)
            if not destination_binary.exists():
                shutil.copy(source_binary, destination_binary)



import glob
import hashlib
import logging
import os
from pathlib import Path

import pycurl
from google.cloud import storage

from helix.configuration import MANAGER_CFG
from helix.progress.base_progress_calculator import BaseProgressCalculator
from helix.reference.genome_metadata_loader import Genome
from helix.files.file_type_checker import FileTypeChecker

logger = logging.getLogger(__name__)
# There libs are very chatty for everything less than WARNING
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("google").setLevel(logging.WARNING)

HANDLERS = {}
SIZE_HANDLERS = {}


def uri_handler(uri):
    def decorator(f):
        global HANDLERS
        HANDLERS[uri] = f
        return f

    return decorator


def size_handler(uri):
    def decorator(f):
        global SIZE_HANDLERS
        SIZE_HANDLERS[uri] = f
        return f

    return decorator


class Downloader:

    def __init__(
        self,
        config=MANAGER_CFG.REPOSITORY,
        file_type_checker: FileTypeChecker = FileTypeChecker(),
    ) -> None:
        if file_type_checker is None:
            raise RuntimeError("FileTypeChecker cannot be None.")

        self._config = config
        self._progressbar = None
        self.file_type_checker = file_type_checker

    def size_pycurl(self, url: str):
        curl = pycurl.Curl()
        curl.setopt(pycurl.URL, url)
        curl.setopt(pycurl.NOBODY, 1)
        curl.perform()
        length = int(curl.getinfo(pycurl.CONTENT_LENGTH_DOWNLOAD))
        if length == -1:
            return None
        if length == 0:
            return None
        return length

    def get_file_size(self, url: str):
        for handler in SIZE_HANDLERS.keys():
            if url.startswith(handler):
                return SIZE_HANDLERS[handler](self, url)
        return self.size_pycurl(url)

    def calculate_md5_hash(self, filename: Path, chunk_size=4096):
        md5_hash = hashlib.md5()
        with filename.open("rb") as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                md5_hash.update(chunk)
        return md5_hash.hexdigest()

    def pre_download_action(self, genome: Genome):
        # Until it's in the temporary directory, the file has
        # a known name but it may have an unknown extension.
        # The extension is assigned after the file is downloaded
        # based on the file content (as the URL can be misleading).
        # As a consequence, the only way to find a file is to
        # look at its size/MD5.

        target = self._config.temporary.joinpath(genome.url_hash)
        if target.exists():
            return target

        files = glob.glob(f"{target}.*")
        for file in files:
            file = Path(file)
            if file.stat().st_size == genome.download_size:
                return file
        return target

    @size_handler("https://storage.cloud.google.com")
    @size_handler("gs://")
    def size_google(self, genome: Genome):
        storage_client = storage.Client.create_anonymous_client()
        uri = genome.fasta_url.strip()
        if not uri.startswith("gs://"):
            uri = "gs://" + uri.replace("https://storage.cloud.google.com/", "")

        blob = storage.Blob.from_string(uri, client=storage_client)
        blob.reload()
        return blob.size

    @uri_handler("https://storage.cloud.google.com")
    @uri_handler("gs://")
    def download_google(self, genome: Genome, callback: any):
        storage_client = storage.Client.create_anonymous_client()
        uri = genome.fasta_url.strip()
        if not uri.startswith("gs://"):
            uri = "gs://" + uri.replace("https://storage.cloud.google.com/", "")

        blob = storage.Blob.from_string(uri, client=storage_client)
        blob.reload()
        if genome.download_size is None:
            genome.download_size = blob.size

        target = self.pre_download_action(genome)

        if target.exists():
            if target.stat().st_size == genome.download_size:
                return self.post_download_action(genome, target)

        blob.download_to_filename(target)
        return self.post_download_action(genome, target)

    def post_download_action(self, genome: Genome, downloaded: Path):
        md5 = self.calculate_md5_hash(downloaded)
        if genome.downloaded_md5 is None:
            genome.downloaded_md5 = md5
        elif genome.downloaded_md5 != md5:
            downloaded.unlink()
            raise RuntimeError(f"MD5 for {genome} is not matching and was deleted.")

        extension = self.file_type_checker.get_extension(downloaded)

        if extension is None:
            _, ext = os.path.splitext(genome.fasta_url)
            extension = ext.lower()

        if extension is None:
            raise RuntimeError(
                f"Format file for {genome} was not recognized. Aborting."
            )

        if downloaded.suffix != extension:
            target = downloaded.with_suffix(extension)
            downloaded.rename(target)
            downloaded = target
        return downloaded

    def perform(self, genome: Genome, progress: any = False) -> Path:
        for handler in HANDLERS.keys():
            if genome.fasta_url.startswith(handler):
                return HANDLERS[handler](self, genome, progress)
        return self.download_pycurl(genome, progress)

    def download_pycurl(self, genome: Genome, progress: any):
        if genome.download_size is None:
            genome.download_size = self.get_file_size(genome.fasta_url)
            if genome.download_size is None:
                raise RuntimeError(f"Unable to download file {genome.fasta_url}")

        target = self.pre_download_action(genome)

        resume_from = None
        if target.exists():
            if target.stat().st_size == genome.download_size:
                return self.post_download_action(genome, target)
            else:
                resume_from = target.stat().st_size

        with target.open("wb" if resume_from is None else "ab") as f:
            curl = pycurl.Curl()
            curl.setopt(pycurl.URL, genome.fasta_url)
            curl.setopt(pycurl.WRITEDATA, f)
            curl.setopt(pycurl.FOLLOWLOCATION, True)
            if progress is not False or progress is not None:
                curl.setopt(pycurl.NOPROGRESS, False)

            if resume_from is not None:
                curl.setopt(pycurl.RESUME_FROM, resume_from)

            # TODO: find an easy way (if it exists) to load ca-bundle
            # on every platform. Or remove this TODO as we don't really
            # care about people MITM-ing our reference genome download.
            curl.setopt(pycurl.SSL_VERIFYPEER, 0)
            curl.setopt(pycurl.SSL_VERIFYHOST, 0)

            progress_calc = BaseProgressCalculator(
                progress, genome.download_size, "Download"
            )

            if progress is not None and progress is not False:
                curl.setopt(
                    pycurl.PROGRESSFUNCTION,
                    lambda tot_down, down, tot_up, up: progress_calc.compute(down),
                )

            curl.perform()
            curl.close()
            # Signal the download is done
            if progress is not None and progress is not False:
                progress(None, None)
        return self.post_download_action(genome, target)
